#! /usr/bin/env python

"""
Learning from interpretations
-----------------------------
Parameter learning for ProbLog.
Given a probabilistic program with parameterized weights and a set of partial implementations,
learns appropriate values of the parameters.

Algorithm
+++++++++
The algorithm operates as follows:
    0. Set initial values for the weights to learn.
    1. Set the evidence present in the example.
    2. Query the model for the weights of the atoms to be learned.
    3. Update the weights to learn by taking the mean value over all examples and queries.
    4. Repeat steps 1 to 4 until convergence (or a maximum number of iterations).
The score of the model for a given example is obtained by calculating the probability of the
evidence in the example.

Implementation
++++++++++++++
The algorithm is implemented on top of the ProbLog toolbox.
It uses the following extensions of ProbLog's classes:
    * a LogicProgram implementation that rewrites the model and extracts the weights to learn
    (see :py:func:`learning.lfi.LFIProblem.__iter__`)
    * a custom semiring that looks up the current value of a weight to learn
    (see :py:func:`learning.lfi.LFIProblem.value`)
.. autoclass:: learning.lfi.LFIProblem
    :members: __iter__, value
"""
import sys
import random
import math
import logging

try:
    from typing import List, Union
except ImportError:
    List, Union = None, None

from collections import defaultdict
from itertools import chain
from problog.util import init_logger
from logging import getLogger
from problog.engine import DefaultEngine, ground
from problog.evaluator import SemiringLogProbability, SemiringProbability
from problog.logic import (
    Term,
    Constant,
    Clause,
    AnnotatedDisjunction,
    Or,
    Var,
    InstantiationError,
    ArithmeticError,
    term2list,
)
from problog.program import PrologString, PrologFile, LogicProgram
from problog.errors import InconsistentEvidenceError, process_error
from problog import get_evaluatable, get_evaluatables
import traceback
import timeit


def str2bool(s):
    if str(s) == "true":
        return True
    elif str(s) == "false":
        return False
    else:
        return None


def str2num(s):
    """Translate a Term that represents a number or list of numbers to observations (as Python primitives).
    :return: Tuple of (isobserved?, values)
    """
    if s.is_constant() and (s.is_float() or s.is_integer()):
        return True, s.compute_value()
    elif s.functor == ".":
        values = term2list(s)
        numvalues = []
        for value in values:
            if isinstance(value, int) or isinstance(value, float):
                numvalues.append(value)
            else:
                return None, None
        return True, tuple(numvalues)
    else:
        return None, None


class LFIProblem(LogicProgram):
    def __init__(
        self,
        source,
        examples,
        max_iter=10000,
        min_improv=1e-10,
        verbose=0,
        knowledge=None,
        leakprob=None,
        propagate_evidence=False,
        infer_AD_values=True,
        normalize=False,
        logspace=False,
        **extra
    ):
        """
        Learn parameters using LFI.
        :param source: filename of file containing input model
        :type source: str
        :param examples: list of observed terms / value
        :type examples: list[list[tuple(Term, bool)]]
        :param max_iter: maximum number of iterations to run
        :type max_iter: int
        :param min_improv: minimum improvement in log-likelihood for convergence detection
        :type min_improv: float
        :param verbose: verbosity level
        :type verbose: int
        :param knowledge: class to use for knowledge compilation
        :type knowledge: class
        :param leakprob: Add all true evidence atoms with the given probability
                         to avoid 'inconsistent evidence' errors. This also
                         allows to learn a program without constants and
                         retrieve the constants from the evidence file.
                         (default: None)
        :type leakprob: float or None
        :param infer_AD_values: enable inferring AD values from given evidence,
                                e.g. for an AD p1::q1; ... ;pn::qn :- b.
                                if b is true and q2,...,qn are all false, then an extra evidence q1 = T is inserted.
                                (see also LFIProblem._process_examples.all_false_except_one)
        :type infer_AD_values: bool
        :param extra: catch all for additional parameters (not used)
        """
        LogicProgram.__init__(self)
        self.source = source
        self._log = logspace

        # The names of the atom for which we want to learn weights.
        self.names = []
        self.bodies = []
        self.parents = []

        # The weights to learn.
        # The initial weights are of type 'float'.
        # When necessary they are replaced by a dictionary [t(arg1, arg2, ...) -> float]
        #  for weights of form t(SV, arg1, arg2, ...).
        self._weights = []

        self.examples = LFIProblem._fix_examples(examples)
        self.leakprob = leakprob
        self.leakprobatoms = None
        self.propagate_evidence = propagate_evidence
        self.infer_AD_values = infer_AD_values
        self._compiled_examples = None

        self.max_iter = max_iter
        self.min_improv = min_improv
        self.verbose = verbose
        self.iteration = 0

        if knowledge is None:
            knowledge = get_evaluatable()
        self.knowledge = knowledge

        self.output_mode = False
        self.extra = extra

        self._enable_normalize = normalize
        self._adatoms = []  # list AD atoms and total probability
        self._adatomc = {}  # complement of AD atom (complement that adds to prob 1.0)
        self._adparent = {}  # atom representing parent of AD

    @staticmethod
    def _fix_examples(examples):
        """ Fixes input format of examples, present in problog 2.1.0.42

            ProbLog 2.1.0.42 contained pieces of code from the work on LFI supporting continuous distributions.
            In this code, self.examples did no longer have the format list[tuple(Term, bool)] but was instead
            storing tuples of size 3, with the third argument being 'cvalue'. This shouldn't have been pushed, and it
            is possible some people have adapted their input format to this. After reverting this format change, this
            method ensures backwards compatibility for the format of tuples with length 3.
            If later this format of length three is restored, remove this method.
        """
        if len(examples) > 0 and len(examples[0]) > 0 and len(examples[0][0]) > 2:
            # Assumes all elements use the same format of tuple of length 3 (so we only check [0][0])
            return [
                [(tuple_ex[0], tuple_ex[1]) for tuple_ex in example]
                for example in examples
            ]
        else:
            return examples

    @property
    def count(self):
        """Number of parameters to learn."""
        return len(self.names)

    def add_ad(self, rem_prob, indices):
        """
        :param rem_prob: Remaining probability that can be learned (if no fixed probabilities given, this will be one)
        :param indices: Indices of atoms that together form an annotated disjunction.
        :return: None
        """
        self._adatoms.append((rem_prob, indices))
        for idx in indices:
            self._adatomc[idx] = [idxo for idxo in indices if idxo != idx]

    def count_ad(self):
        return len(self._adatoms)

    def append_ad(self, atom_index, ad_index=None):
        if ad_index is None:
            ad_index = -1
        self._adatoms[ad_index][1].append(atom_index)
        indices = self._adatoms[ad_index][1]
        for idx in indices:
            self._adatomc[idx] = [idxo for idxo in indices if idxo != idx]

    def verify_ad(self, ad_index=None):
        if ad_index is None:
            ad_index = -1
        if len(self._adatoms[ad_index][1]) == 1:
            indices = self._adatoms[ad_index][1]

            self._adatomc[indices[0]] = [
                -1 - indices[0]
            ]  # No AD, complement is negative variable

    def prepare(self):
        """Prepare for learning."""
        self._compile_examples()

    def _get_weight(self, index, args, strict=True):
        index = int(index)
        weight = self._weights[index]
        if isinstance(weight, dict):
            if strict:
                return weight[args]
            else:
                return weight.get(args, 0.0)
        else:
            return weight

    def get_weights(self, index):
        """Get a list of key, weight pairs for the given input fact.
        :param index: identifier of the fact
        :return: list of key, weight pairs where key refers to the additional variables
        on which the weight is based
        :rtype: list[tuple[Term, float]]
        """
        weight = self._weights[index]
        if isinstance(weight, dict):
            return list(weight.items())
        else:
            return [(Term("t"), weight)]

    def _set_weight(self, index, args, weight, weight_changed=None):
        index = int(index)
        if not args:
            # assert not isinstance(self._weights[index], dict)
            self._weights[index] = weight
        elif isinstance(self._weights[index], dict):
            if (
                self._weights
                and self._weights[index]
                and weight_changed
                and args in weight_changed[index]
                and weight_changed[index][args]
            ):
                self._weights[index][args] += weight
            else:
                self._weights[index][args] = weight
        else:
            self._weights[index] = {args: weight}

    def _add_weight(self, weight):
        self._weights.append(weight)

    def _process_examples(self):
        """Process examples by grouping together examples with similar structure.
        :return: example groups based on evidence atoms
        :rtype: dict of atoms : values for examples
        """
        logger = getLogger("problog_lfi")
        # value can be True / False / None
        # ( atom ), ( ( value, ... ), ... )

        # Simple implementation: don't add neutral evidence.

        # ad_groups is a list of lists where each list contains an AD
        ad_groups = list()
        for ad in self._adatoms:
            # if it's an AD group
            if len(ad[1]) > 1:
                ad_list = []
                for var in ad[1]:
                    ad_list.append(Term(self.names[var].functor, *self.names[var].args))
                ad_groups.append(tuple(ad_list))
        logger.debug("AD Groups\t\t:" + str(ad_groups))

        def all_false(d):
            """
            This function recognizes inconsistent evidence s.t. all values are False in AD.
            :param d: dictionary of ADs in form {term: value}
                    value can be True, False, None, "Template"
            :return: whether all values are False
            """
            # false_count should be the same as the length of d
            false_count = sum(v is False for v in d.values())
            return false_count == len(d)

        def all_false_except_one(d):
            """
            This function recognizes incomplete evidence s.t.
            the non-False value in AD needs to be set to True.
            :param d: dictionary of ADs in form {term: value}
                    value can be True, False, None, "Template"
            :return: whether all values except one are False
            """
            false_count = sum(v is False for v in d.values())
            the_left_is_none = bool(sum(v is None for v in d.values()))
            return (false_count == len(d) - 1) and the_left_is_none

        def getADtemplate(d, atom=None):
            """
            This function gets atom's complement AD template.
            This should only be used when the AD contains non-ground terms.
            :param d: dictionary of ADs in form {term: value}
                    value can be True, False, None, "Template"
            :param atom: an evidence
            :return: atom's complement AD template
            """
            if atom is not None:
                temp_dict = {
                    k: v
                    for k, v in d.items()
                    if v == "Template" and atom.signature != k.signature
                }
                return temp_dict
            else:
                temp_dict = {k: v for k, v in d.items() if v == "Template"}
                return temp_dict

        def add_to_ad_evidence(pair, l, ADtemplate):
            """
            :param pair: a new pair of (atom, value)
            :param l: a list of dictionaries, all dictionaries need to have the same format
            :return:
            """
            (k, v) = pair
            # if entry k exists, update the value with k
            for d in l:
                if k in d:
                    d[k] = v
                    return
            # if entry k does not exist, create a new dictionary from template
            # and instantiate it with k
            new_d = dict()
            for temp_k in ADtemplate.keys():
                new_key = Term(temp_k.functor, *k.args)
                new_d[new_key] = None
            # put v in there
            new_d[k] = v
            l.append(new_d)

        if self.infer_AD_values:
            result = ExampleSet()
            inconsistent = False
            # iterate over all examples given in .ev
            for index, example in enumerate(self.examples):
                ad_evidences = []
                non_ad_evidence = {}
                for ad_group in ad_groups:
                    # create a dictionary to memorize what evidence is given in AD
                    d = dict()
                    # TODO: what if the AD contains both ground and non-ground????
                    # e.g. t(_)::a; t(_)::b(X)
                    for var in ad_group:
                        if var.is_ground():
                            d[var] = None  # for ground unknown evidence
                        else:
                            d[var] = "Template"  # for unground unknown evidence
                    ad_evidences.append(d)

                # add all evidence in the example to ad_evidences
                for atom, value in example:
                    # if atom has a tunable probability to learn
                    if any([atom.signature == name.signature for name in self.names]):
                        if len(atom.args) == 0: # Propositional evidence
                            # insert evidence
                            is_ad = False
                            for d in ad_evidences:
                                if atom in d:
                                    d[atom] = value
                                    is_ad = True
                            if not is_ad:
                                non_ad_evidence[
                                    atom
                                ] = value
                        else: # First Order evidence
                            # find the right AD dictionary : AD_dict
                            AD_dict = None
                            for d in ad_evidences:
                                if any([atom.signature == k.signature for k in d]):
                                    AD_dict = d
                            # if the instantiation is new, add it as a key to the dictionary
                            if AD_dict and AD_dict.get(atom) is None:
                                AD_dict[atom] = value
                                # also add other AD parts in the dictionary with value==None
                                other_ADs = getADtemplate(AD_dict, atom)
                                for otherAD in other_ADs.keys():
                                    new_key = Term(otherAD.functor, *atom.args)
                                    AD_dict[new_key] = AD_dict.get(new_key, None)
                            else:
                                non_ad_evidence[atom] = value
                    else:
                        non_ad_evidence[atom] = value

                # grounded_ad_evidences contains all usable evidence (gound, not template)
                grounded_ad_evidences = []
                for d in ad_evidences:
                    # for first order evidence dictionaries
                    if "Template" in d.values():
                        # new_ad_evidence is a list of dictionaries
                        # each dictionary is a group of the AD template instantiation
                        new_ad_evidence = list()
                        # get template AD evidence
                        ADtemplate = getADtemplate(d)
                        # group all pairs according to ADtemplate
                        for k, v in d.items():
                            if v != "Template":
                                add_to_ad_evidence((k, v), new_ad_evidence, ADtemplate)
                        grounded_ad_evidences += new_ad_evidence
                    else:
                        # simply us them
                        grounded_ad_evidences.append(d)

                inconsistent_example = False
                for i, d in enumerate(grounded_ad_evidences):
                    # inconsistent1 = multiple_true(d)
                    inconsistent2 = all_false(d)
                    add_compliment = all_false_except_one(d)

                    if inconsistent2:
                        inconsistent_example = True
                        continue
                    elif add_compliment:
                        for key, value in d.items():
                            if value is None:
                                grounded_ad_evidences[i][key] = True

                if not inconsistent_example and len(grounded_ad_evidences) > 0:
                    # There are (fully tunable) ADs in the program
                    evidence_list = []
                    for d in grounded_ad_evidences:
                        for key, value in d.items():
                            if value is not None:
                                if (key, value) not in evidence_list:
                                    # TODO: Switch to orderedSet so we don't have to check whether it's in already
                                    evidence_list.append((key, value))

                    for key, value in non_ad_evidence.items():
                        evidence_list.append((key, value))

                    atoms, values = zip(*evidence_list)
                    result.add(index, atoms, values)

                else:
                    # (No AD case) or (Inconsistent Evidence Case)
                    atoms, values = zip(*example)
                    result.add(index, atoms, values)
            return result
        else:
            # smarter: compile-once all examples with same atoms
            result = ExampleSet()
            for index, example in enumerate(self.examples):
                atoms, values = zip(*example)
                result.add(index, atoms, values)
            return result

    def _compile_examples(self):
        """Compile examples."""
        logger = getLogger("problog_lfi")
        baseprogram = DefaultEngine(**self.extra)
        baseprogram = baseprogram.prepare(self)
        logger.debug(
            "\nBase Program:\n\t" + baseprogram.to_prolog().replace("\n", "\n\t")
        )
        examples = self._process_examples()
        for i, example in enumerate(examples):
            logger.debug("\nCompiling example {}/{}".format(i + 1, len(examples)))
            example.compile(self, baseprogram)
        self._compiled_examples = examples
        
        # for example in examples:
        #     logger.info(
        #         f"Circuit size: \n\t{example.n}\n\t{len(example.compiled)}"
        #     )

    def _process_atom(self, atom, body):
        """Returns tuple ( prob_atom, [ additional clauses ] )"""
        if isinstance(atom, Or):
            # Annotated disjunction
            atoms = atom.to_list()
        else:
            atoms = [atom]

        atoms_out = []
        extra_clauses = []

        has_lfi_fact = False
        prior_probability = 0.0  # Sum of prior weights in AD.
        fixed_probability = 0.0  # Sum of fixed (i.e. non-learnable) weights in AD.

        num_random_weights = 0
        for atom in atoms:
            if atom.probability and atom.probability.functor == "t":
                try:
                    start_value = float(atom.probability.args[0])
                    prior_probability += float(start_value)
                except InstantiationError:
                    # Can't be converted to float => take random
                    num_random_weights += 1
                except ArithmeticError:
                    num_random_weights += 1
            elif atom.probability and atom.probability.is_constant():
                fixed_probability += float(atom.probability)

        random_weights = [random.random() for _ in range(0, num_random_weights + 1)]
        norm_factor = (1.0 - prior_probability - fixed_probability) / sum(
            random_weights
        )
        random_weights = [r * norm_factor for r in random_weights]

        # First argument is probability available for learnable weights in the AD.
        self.add_ad(1.0 - fixed_probability, [])  # TODO : this adds extra ad

        # Replace anonymous variables with non-anonymous variables.
        class ReplaceAnon(object):
            def __init__(self):
                self.cnt = 0

            def __getitem__(self, key):
                if key == "_":
                    self.cnt += 1
                    return Var("anon_%s" % self.cnt)
                else:
                    return Var(key)

        prob_args = []
        if isinstance(atom.probability, Term):
            for arg in atom.probability.args:
                if not isinstance(arg, Constant) and arg != Var("_"):
                    prob_args.append(arg)

        newcount = "_".join([str(self.count + count) for count in range(len(atoms))])

        factargs = []
        # probargs = ()
        for atom in atoms:
            q = list(atom.apply(ReplaceAnon()).args)
            for var in q:
                if var not in factargs:
                    factargs.append(var)

        lfi_rule = Term(
            "lfi_rule", Constant(newcount), Term("t", *prob_args), *factargs
        )

        if body is not None:
            extra_clauses.append(Clause(lfi_rule, body))

        for atom in atoms:
            if atom.probability and atom.probability.functor == "t":
                # t(_, X)::p(X, Y) :- body.
                #
                # Translate to
                #     lfi_prob(1, t(X))::lfi_fact(1, t(X), X, Y).
                #     lfi_rule(1, t(X), X, Y): - body.
                #     p(X, Y): - lfi_body(1, t(X), X, Y).
                #     lfi_body(1, t(X), X, Y): - lfi_par(1, t(X), X, Y), lfi_fact(1, t(X), X, Y).
                #     lfi_par(1, t(X), X, Y): - lfi_rule(1, t(X), X, Y).
                #
                #
                # For annotated disjunction: t(_)::p1(X); t(_)::p2(X): - body.
                #     lfi_prob(0, t)::lfi_fact(0, t, X); lfi_prob(1, t)::lfi_fact(1, t, X): - lfi_rule(0_1, t, X).
                #     lfi_rule(0_1, t, X): - body.
                #     p1(X): - lfi_body(0, t, X).
                #     lfi_body(0, t, X): - lfi_par(0, t, X), lfi_fact(0, t, X).
                #     lfi_par(0, t, X): - lfi_rule(0_1, t, X).
                #     p2(X): - lfi_body(1, t, X).
                #     lfi_body(1, t, X): - lfi_par(1, t, X), lfi_fact(1, t, X).
                #     lfi_par(1, t, X): - lfi_rule(0_1, t, X).
                #     ...
                has_lfi_fact = True
                # Learnable probability
                try:
                    start_value = float(atom.probability.args[0])
                except InstantiationError:
                    start_value = None
                except ArithmeticError:
                    start_value = None

                atom1 = atom.apply(ReplaceAnon())

                # 1) Introduce a new LFI terms
                factargs = atom1.args
                lfi_fact = Term(
                    "lfi_fact", Constant(self.count), Term("t", *prob_args), *factargs
                )
                lfi_body = Term(
                    "lfi_body", Constant(self.count), Term("t", *prob_args), *factargs
                )
                lfi_par = Term(
                    "lfi_par", Constant(self.count), Term("t", *prob_args), *factargs
                )
                lfi_prob = Term("lfi_prob", Constant(self.count), Term("t", *prob_args))

                # 2) Replacement atom
                replacement = lfi_fact.with_probability(lfi_prob)

                # 3) Create redirection clause
                extra_clauses.append(Clause(atom1.with_probability(), lfi_body))
                extra_clauses.append(Clause(lfi_body, lfi_par & lfi_fact))

                if body is None:
                    extra_clauses.append(Clause(lfi_par, Term("true")))
                else:
                    extra_clauses.append(Clause(lfi_par, lfi_rule))

                self.append_ad(len(self._weights))

                # 4) Set initial weight
                if start_value is None:
                    # Assign a random weight initially
                    start_value = random_weights.pop(-1)
                self._add_weight(start_value)

                # 5) Add name
                self.names.append(atom)
                self.bodies.append(lfi_body)
                self.parents.append(lfi_par)
                atoms_out.append(replacement)
            else:
                atoms_out.append(atom)

        self.verify_ad()

        if has_lfi_fact:
            if len(atoms) == 1 and body is None:
                # Non AD
                return [atoms_out[0]] + extra_clauses
            else:
                # AD
                if body is None:
                    return [
                        AnnotatedDisjunction(atoms_out, Term("true"))
                    ] + extra_clauses
                else:
                    return [AnnotatedDisjunction(atoms_out, lfi_rule)] + extra_clauses
        else:
            if len(atoms) == 1:
                if body is None:
                    return [atoms_out[0]]
                else:
                    return [Clause(atoms_out[0], body)]
            else:
                if body is None:
                    body = Term("true")
                return [AnnotatedDisjunction(atoms_out, body)]

    def _process_atom_output(self, atom, body):
        """Returns tuple ( prob_atom, [ additional clauses ] )"""

        if isinstance(atom, Or):
            atoms = atom.to_list()
        else:
            atoms = [atom]

        transforms = defaultdict(list)

        atoms_fixed = []
        fixed_only = True
        for atom in atoms:
            if atom.probability and atom.probability.functor == "t":
                assert atom in self.names

                index = self.output_names.index(atom)
                weights = self.get_weights(index)

                for w_args, w_val in weights:
                    translate = tuple(zip(atom.probability.args[1:], w_args.args))
                    transforms[translate].append(atom.with_probability(Constant(w_val)))
                self.output_names[index] = None
                fixed_only = False
            else:
                atoms_fixed.append(atom)

        if not fixed_only:
            clauses = []
            for tr, atoms in transforms.items():
                tr = DefaultDict({k: v for k, v in tr})
                atoms_out = [at.apply(tr) for at in atoms] + atoms_fixed
                if len(atoms_out) == 1:
                    if body is None:
                        clauses.append(atoms_out[0])
                    else:
                        clauses.append(Clause(atoms_out[0], body.apply(tr)))
                else:
                    if body is None:
                        clauses.append(AnnotatedDisjunction(atoms_out, None))
                    else:
                        clauses.append(AnnotatedDisjunction(atoms_out, body.apply(tr)))
            return clauses
        else:
            atoms_out = atoms_fixed
            if len(atoms_out) == 1:
                if body is None:
                    return [atoms_out[0]]
                else:
                    return [Clause(atoms_out[0], body)]
            else:
                return [AnnotatedDisjunction(atoms_out, body)]

    # Overwrite from LogicProgram
    def __iter__(self):
        """
        Iterate over the clauses of the source model.
        This object can be used as a LogicProgram to be passed to the grounding Engine.
        Extracts and processes all ``t(...)`` weights.
        This
            * replaces each probabilistic atom ``t(_, X)::p(X, Y) :- b(X, Y)`` by a unique atom \
            ``lfi_prob(i,t(X))::lfi_fact(i,t(X),X,Y)``;
            * adds a new clause ``p(X,Y) :- lfi_body(1,t(X),X,Y)``;
            * adds a new clause ``lfi_body(i,t(X),X,Y) :- lfi_par(i,t(X),X,Y), lfi_fact(i,t(X),X,Y)``;
            * adds a new clause ``lfi_par(i,t(X),X,Y) :- lfi_rule(i,t(X),X,Y)``;
            * adds a new clause ``lfi_rule(i,t(X),X,Y) :- b(X,Y)``;
            * initializes the weight of ``lfi_prob(i,t(X))`` based on the ``t(...)`` specification;
        This also removes all existing queries from the model.
        Example:
        .. code-block:: prolog
            t(_) :: b(X).
            t(_) :: p(X) :- b(X).
        is transformed into
        .. code-block:: prolog
            lfi_prob(0,t)::lfi_fact(0,t,X).
            b(X) :- lfi_body(0,t,X).
            lfi_body(0,t,X) :- lfi_par(0,t,X), lfi_fact(0,t,X).
            lfi_par(0,t,X) :- true.
            lfi_prob(1,t)::lfi_fact(1,t,X).
            lfi_rule(1,t,X) :- b(X).
            p(X) :- lfi_body(1,t,X).
            lfi_body(1,t,X) :- lfi_par(1,t,X), lfi_fact(1,t,X).
            lfi_par(1,t,X) :- lfi_rule(1,t,X).

        If ``self.leakprobs`` is a value, then during learning all true
        examples are added to the program with the given leak probability.
        """

        if self.output_mode:
            process_atom = self._process_atom_output
            self.output_names = self.names[:]
        else:
            process_atom = self._process_atom

        if self.output_mode is False:
            getLogger("problog_lfi").debug("\nProcessed Atoms:")
        for clause in self.source:
            if isinstance(clause, Clause):
                if clause.head.functor == "query" and clause.head.arity == 1:
                    continue
                extra_clauses = process_atom(clause.head, clause.body)
                for extra in extra_clauses:
                    if self.output_mode is False:
                        getLogger("problog_lfi").debug("\t" + str(extra))
                    yield extra
            elif isinstance(clause, AnnotatedDisjunction):
                extra_clauses = process_atom(Or.from_list(clause.heads), clause.body)
                for extra in extra_clauses:
                    if self.output_mode is False:
                        getLogger("problog_lfi").debug("\t" + str(extra))
                    yield extra
            else:
                if clause.functor == "query" and clause.arity == 1:
                    continue
                # Fact
                extra_clauses = process_atom(clause, None)
                for extra in extra_clauses:
                    if self.output_mode is False:
                        getLogger("problog_lfi").debug("\t" + str(extra))
                    yield extra

        if self.leakprob is not None:
            leakprob_atoms = self._get_leakprobatoms()
            for example_atom in leakprob_atoms:
                yield example_atom.with_probability(Constant(self.leakprob))

    def _get_leakprobatoms(self):
        if self.leakprobatoms is not None:
            return self.leakprobatoms
        self.leakprobatoms = set()
        for examples in self.examples:
            for example, obs in examples:
                if obs:
                    self.leakprobatoms.add(example)
        return self.leakprobatoms

    def _evaluate_examples(self):
        """Evaluate the model with its current estimates for all examples."""

        getLogger("problog_lfi").debug("Evaluating examples:")
        if self._log:
            evaluator = ExampleEvaluatorLog(self._weights)
        else:
            evaluator = ExampleEvaluator(self._weights)

        results = []
        for i, example in enumerate(self._compiled_examples):
            try:
                result = evaluator(example)
                results.append(result)
                getLogger("problog_lfi").debug(
                    "Example "
                    + str(i + 1)
                    + ":\tFrequency = "
                    + str(result[0][0])
                    + "\tp_evidence = "
                    + str(result[0][1])
                    + "\tp_queries = "
                    + str(result[0][2])
                )
            except InconsistentEvidenceError:
                # print("Ignoring example {}/{}".format(i + 1, len(self._compiled_examples)))
                getLogger("problog_lfi").warning(
                    "Ignoring example {}/{}".format(i + 1, len(self._compiled_examples))
                )

        return list(chain.from_iterable(results))

    def _update(self, results):
        """Update the current estimates based on the latest evaluation results."""
        logger = getLogger("problog_lfi")
        fact_body = defaultdict(int)
        fact_par = defaultdict(int)

        log_likelihood = 0.0
        for m, pEvidence, result in results:
            par_marg = dict()
            for fact, value in result.items():
                # use the id and the t variables as index
                index = (fact.args[0], fact.args[1])
                if fact.functor == "lfi_body":
                    fact_body[index] += value * m
                elif fact.functor == "lfi_par":
                    if index in par_marg:
                        par_marg[index] += value
                        for o_index in self._adatomc[index[0]]:
                            par_marg[(o_index, *index[1:])] += value
                    else:
                        par_marg[index] = value
                        for o_index in self._adatomc[index[0]]:
                            par_marg[(o_index, *index[1:])] = value

            for index, value in par_marg.items():
                fact_par[index] += value * m
            try:
                log_likelihood += m * math.log(pEvidence)
            except ValueError:
                logger.debug("Pr(evidence) == 0.0")

        update_list = fact_body

        weight_changed = []
        for weight in self._weights:
            if isinstance(weight, float):
                weight_changed.append(False)
            elif isinstance(weight, dict):
                d = dict()
                for w in weight:
                    d[w] = False
                weight_changed.append(d)

        convergence_score = 0.0
        for index in update_list:
            if float(fact_body[index]) <= 10**-15:
                # if close to zero
                prob = 0.0
            else:
                prob = float(fact_body[index]) / float(fact_par[index])
                try:
                    convergence_score += math.log(prob)
                except ValueError as ex:
                    # prob too close to zero
                    pass

            logger.debug(
                "Update probabilistic fact {}: {} / {} = {}".format(
                    index, fact_body[index], fact_par[index], prob
                )
            )
            self._set_weight(index[0], index[1], prob, weight_changed=weight_changed)
            if not index[1]:
                weight_changed[int(index[0])] = True
            elif (
                isinstance(weight_changed[int(index[0])], dict)
                and index[1] in weight_changed[int(index[0])]
            ):
                weight_changed[int(index[0])][index[1]] = True
            else:
                weight_changed[int(index[0])] = {index[1]: True}

        if self._enable_normalize:
            self._normalize_weights()

        return log_likelihood, convergence_score

    def _normalize_weights(self):
        # TODO: too late here, AD should be taken into account in _update
        # Derivation is sum(all values for var=k) / sum(all values for i sum(all values for var=i))

        for available_prob, idx in self._adatoms:
            if len(idx) == 1:
                # Not an AD; No need to normalize
                continue
            keys = set()
            for i in idx:
                for key, val in self.get_weights(i):
                    keys.add(key)

            keys = list(keys)
            for key in keys:
                w = sum(self._get_weight(i, key, strict=False) for i in idx)
                if w != 0:
                    n = (
                        available_prob / w
                    )  # Some part of probability might be taken by non-learnable weights in AD.
                else:
                    n = available_prob

                for i in idx:
                    self._set_weight(
                        i, key, self._get_weight(i, key, strict=False) * n,
                    )

    def step(self):
        self.iteration += 1
        getLogger("problog_lfi").info("\nIteration " + str(self.iteration))
        results = self._evaluate_examples()
        return self._update(results)

    def get_model(self):
        self.output_mode = True
        lines = []
        for l in self:
            lines.append("%s." % l)
        lines.append("")
        self.output_mode = False
        return "\n".join(lines)

    def run(self):
        tic = timeit.default_timer()
        self.prepare()
        toc = timeit.default_timer()
        getLogger("problog_lfi").info("Compile time: %f" % (toc - tic))
        getLogger("problog_lfi").info("Weights to learn: %s" % self.names)
        getLogger("problog_lfi").info("Bodies: %s" % self.bodies)
        getLogger("problog_lfi").info("Parents: %s" % self.parents)
        getLogger("problog_lfi").info("Initial weights: %s" % self._weights)

        # if self._log:
        #     new_w = []
        #     for w in self._weights:
        #         new_w.append(math.log(w))
        #     self._weights = new_w

        delta = 1000
        prev_score = -1e10
        log_likelihood = 0.0
        tic = timeit.default_timer()
        # while self.iteration < self.max_iter and (delta > self.min_improv):
        while self.iteration < self.max_iter and (delta < 0 or delta > self.min_improv):
            log_likelihood, convergence_score = self.step()
            getLogger("problog_lfi").info(
                "Weights after iteration %s: %s" % (self.iteration, self._weights)
            )
            getLogger("problog_lfi").info(
                "Score after iteration %s: %s" % (self.iteration, log_likelihood)
            )
            # Let p_1, ..., p_N the parameters we need to learn
            # Let p_{1,m}, ..., p_{n,m} the estimates in the m-th iteration
            # delta is (log(p_{1,m+1}) + ... + log(p_{N,m+1})) - (log(p_{1,m}) + ... + log(p_{N,m}))
            # In other words, 10^delta = (p_{1,m+1} * ... *p_{N,m+1})/(p_{1,m} * ... * p_{N,m})
            # delta = abs(score - prev_score)
            delta = convergence_score - prev_score
            prev_score = convergence_score
        toc = timeit.default_timer()
        getLogger("problog_lfi").info("Evaluation time: %f" % (toc - tic))
        return log_likelihood


class ExampleSet(object):
    def __init__(self):
        self._examples = {}

    def add(self, index, atoms, values):
        ex = self._examples.get((atoms, values))
        if ex is None:
            self._examples[(atoms, values)] = Example(index, atoms, values)
        else:
            ex.add_index(index)

    def __iter__(self):
        return iter(self._examples.values())

    def __len__(self):
        return len(self._examples)


class Example(object):
    def __init__(self, index, atoms, values):
        """An example consists of a list of atoms and their corresponding values (True/False).
        Different indices are stored in self.indices
        """
        self.atoms = tuple(atoms)
        self.values = tuple(values)
        self.compiled = []
        self.n = [index]

    def __hash__(self):
        return hash((self.atoms, self.values))

    def __eq__(self, other):
        if other is None:
            return False
        return self.atoms == other.atoms and self.values == other.values

    def compile(self, lfi, baseprogram):
        logger = getLogger("problog_lfi")
        ground_program = None  # Let the grounder decide
        logger.debug("\tGrounded Atoms:\t" + str(self.atoms))
        logger.debug("\tEvidence:\t" + str(list(zip(self.atoms, self.values))))

        ground_program = ground(
            baseprogram,
            ground_program,
            evidence=list(zip(self.atoms, self.values)),
            propagate_evidence=lfi.propagate_evidence,
        )

        logger.debug(
            "\t"
            + "New ground_program:\n\t\t"
            + str(ground_program).replace("\n", "\n\t\t")
        )

        lfi_queries = set()
        for i, node, t in ground_program:
            if (
                t == "atom"
                and isinstance(node.probability, Term)
                and node.probability.functor == "lfi_prob"
            ):
                factargs = ()
                probargs = ()
                if node.name.functor != "choice":
                    if node.name.functor == "lfi_fact":
                        factargs = node.name.args[2:]
                    elif node.name.functor == "lfi_body" or node.name.functor == "lfi_par":
                        factargs = node.name.args[2:]
                        probargs = node.probability.args[1].args
                    else:
                        factargs = node.name.args
                        probargs = node.probability.args[1].args
                elif type(node.identifier) == tuple:
                    factargs = node.name.args[2].args[2:]
                    probargs = node.probability.args[1].args
                lfi_queries.add((node.probability.args[0], probargs, factargs))

        ats = lfi._adatoms
        def add_queries(old_lfi_queries):
            new_lfi_queries = set()
            for _, ad_group in ats:
                for ad in ad_group:
                    if ad in [i[0] for i in old_lfi_queries]:
                        for old_lfi_query in [i for i in old_lfi_queries if i[0] == ad]:
                            for other_ad in ad_group:
                                new_lfi_queries.add((other_ad, old_lfi_query[1], old_lfi_query[2]))
            return new_lfi_queries
            
        # add all AD heads to the query set until no more queries can be added
        old_lfi_queries = set()
        new_lfi_queries = lfi_queries
        
        while old_lfi_queries != new_lfi_queries:
            old_lfi_queries = new_lfi_queries
            new_lfi_queries = add_queries(old_lfi_queries)

        lfi_queries = []
        for q in new_lfi_queries:
            tmp_body = Term("lfi_body", Constant(q[0]), Term("t", *q[1]), *q[2])
            lfi_queries.append(tmp_body)
            logger.debug(
                "\tNode "
                # + str(i)
                + ":\tAdding query for body:\t"
                + str(tmp_body)
                + "\t"
            )
            tmp_par = Term("lfi_par", Constant(q[0]), Term("t", *q[1]), *q[2])
            lfi_queries.append(tmp_par)
            logger.debug(
                "\tNode "
                # + str(i)
                + ":\tAdding query for par :\t"
                + str(tmp_par)
                + "\t"
            )

        ground_program = ground(
            baseprogram,
            ground_program,
            evidence=list(zip(self.atoms, self.values)),
            propagate_evidence=lfi.propagate_evidence,
            queries=lfi_queries,
        )
        logger.debug(
            "\t"
            + "New ground_program:\n\t\t"
            + str(ground_program).replace("\n", "\n\t\t")
        )

        self.compiled = lfi.knowledge.create_from(ground_program)
        try:
            logger.debug(
                "\tCompiled program:\n\t\t"
                + self.compiled.to_prolog().replace("\n", "\n\t\t")
            )
        except Exception:
            logger.debug(
                "\tCompiled program:\n\t\t" + str(self.compiled).replace("\n", "\n\t\t")
            )

    def add_index(self, index):
        self.n.append(index)


class ExampleEvaluator(SemiringProbability):
    def __init__(self, weights):
        SemiringProbability.__init__(self)
        self._weights = weights

    def _get_weight(self, index, args, strict=True):
        index = int(index)
        weight = self._weights[index]
        if isinstance(weight, dict):
            if strict:
                return weight[args]
            else:
                return weight.get(args)
        else:
            return weight

    def is_dsp(self):
        """Indicates whether this semiring requires solving a disjoint sum problem."""
        return True

    def in_domain(self, a):
        return True  # TODO implement

    def value(self, a):
        """Overrides from SemiringProbability.
        Replaces a weight of the form ``lfi_prob(i, t(...))`` by its current estimated value.
        Other weights are passed through unchanged.
        :param a: term representing the weight
        :type a: Term
        :return: current weight
        :rtype: float
        """
        if isinstance(a, Term) and a.functor == "lfi_prob":
            # index = int(a.args[0])
            return self._get_weight(*a.args)
        else:
            return float(a)

    def __call__(self, example):
        """Evaluate the model with its current estimates for all examples."""
        return [
            self._call_internal(
                example.atoms, example.values, example.compiled, example.n
            )
        ]

    def _call_internal(self, at, val, comp, n):

        evidence = {}
        for a, v in zip(at, val):
            if a in evidence:
                if evidence[a] != v:
                    context = " (found evidence({},{}) and evidence({},{}) in example {})".format(
                        a,
                        evidence[a],
                        a,
                        v,
                        ",".join([str(ni) for ni in n])
                        if isinstance(n, list)
                        else n + 1,
                    )
                    raise InconsistentEvidenceError(source=a, context=context)
            else:
                evidence[a] = v

        try:
            # TODO: The next step generates the entire formula if it is density and this is redone later (caching?)
            evaluator = comp.get_evaluator(semiring=self, evidence=evidence)
        except InconsistentEvidenceError as err:
            n = ",".join([str(ni + 1) for ni in n]) if isinstance(n, list) else n + 1
            context = err.context + " (example {})".format(n)
            raise InconsistentEvidenceError(err.source, context)

        p_queries = {}
        # Probability of query given evidence
        for name, node, label in evaluator.formula.labeled():
            if name.functor not in ["lfi_body", "lfi_par"]:
                continue
            w = evaluator.evaluate(node)
            if w < 1e-6:
                p_queries[name] = 0.0
            else:
                p_queries[name] = w
        p_evidence = evaluator.evaluate_evidence()

        return len(n), p_evidence, p_queries


class ExampleEvaluatorLog(SemiringLogProbability):
    def __init__(self, weights):
        SemiringLogProbability.__init__(self)
        self._weights = weights

    def _get_weight(self, index, args, strict=True):
        index = int(index)
        weight = self._weights[index]
        if isinstance(weight, dict):
            if strict:
                weight = weight[args]
            else:
                weight = weight.get(args, 0.0)
        try:
            result = math.log(weight)
        except ValueError:
            return float("-inf")
        return result

    def value(self, a):
        """Overrides from SemiringProbability.
        Replaces a weight of the form ``lfi(i, t(...))`` by its current estimated value.
        Other weights are passed through unchanged.
        :param a: term representing the weight
        :type a: Term
        :return: current weight
        :rtype: float
        """
        if isinstance(a, Term) and a.functor == "lfi_prob":
            rval = self._get_weight(*a.args)
        else:
            rval = math.log(float(a))
        return rval

    def __call__(self, example):
        """Evaluate the model with its current estimates for all examples."""
        return [
            self._call_internal(
                example.atoms, example.values, example.compiled, example.n
            )
        ]

    def _call_internal(self, at, val, comp, n):

        evidence = {}

        for a, v in zip(at, val):
            if a in evidence:
                if evidence[a] != v:
                    context = " (found evidence({},{}) and evidence({},{}) in example {})".format(
                        a,
                        evidence[a],
                        a,
                        v,
                        ",".join([str(ni) for ni in n])
                        if isinstance(n, list)
                        else n + 1,
                    )
                    raise InconsistentEvidenceError(source=a, context=context)
            else:
                evidence[a] = v

        try:
            # TODO: The next step generates the entire formula if it is density and this is redone later (caching?)
            evaluator = comp.get_evaluator(semiring=self, evidence=evidence)
        except InconsistentEvidenceError as err:
            n = ",".join([str(ni + 1) for ni in n]) if isinstance(n, list) else n + 1
            context = err.context + " (example {})".format(n)
            raise InconsistentEvidenceError(err.source, context)

        p_queries = {}
        # Probability of query given evidence
        for name, node, label in evaluator.formula.labeled():
            if name.functor not in ["lfi_body", "lfi_par"]:
                continue
            w = evaluator.evaluate(node)

            p_queries[name] = w

        p_evidence = evaluator.evaluate_evidence()
        # p_evidence = math.exp(p_evidence)
        return len(n), p_evidence, p_queries


def extract_evidence(pl):
    engine = DefaultEngine()
    atoms = engine.query(pl, Term("evidence", None, None))
    atoms1 = engine.query(pl, Term("evidence", None))
    atoms2 = engine.query(pl, Term("observe", None))
    for atom in atoms1 + atoms2:
        atom = atom[0]
        if atom.is_negated():
            atoms.append((-atom, Term("false")))
        else:
            atoms.append((atom, Term("true")))
    result = []
    for at, vl in atoms:
        vlr = str2bool(vl)
        result.append((at, vlr))
    return result


def read_examples(*filenames):
    for filename in filenames:
        engine = DefaultEngine()

        with open(filename) as f:
            example = ""
            for line in f:
                if line.strip().startswith("---"):
                    pl = PrologString(example)
                    atoms = extract_evidence(pl)
                    if len(atoms) > 0:
                        yield atoms
                    example = ""
                else:
                    example += line
            if example:
                pl = PrologString(example)
                atoms = extract_evidence(pl)
                if len(atoms) > 0:
                    yield atoms


class DefaultDict(object):
    def __init__(self, base):
        self.base = base

    def __getitem__(self, key):
        return self.base.get(key, Var(key))


def run_lfi(program, examples, output_model=None, **kwdargs):
    lfi = LFIProblem(program, examples, **kwdargs)
    score = lfi.run()

    getLogger("problog_lfi").info("\nLearned Model:\t\n" + lfi.get_model())
    if output_model is not None:
        with open(output_model, "w") as f:
            f.write(lfi.get_model())

    names = []
    weights = []
    for i, name in enumerate(lfi.names):
        weights_i = lfi.get_weights(i)
        for w_args, w_val in weights_i:
            translate = {k: v for k, v in zip(name.probability.args[1:], w_args.args)}
            names.append(name.apply(DefaultDict(translate)))
            weights.append(w_val)

    return score, weights, names, lfi.iteration, lfi


def argparser():
    import argparse

    parser = argparse.ArgumentParser(
        description="Learning from interpretations with ProbLog"
    )
    parser.add_argument("model")
    parser.add_argument("examples", nargs="+")
    parser.add_argument(
        "-s",
        dest="seed",
        default=None,
        type=int,
        help="The seed to use when sampling initial values",
    )
    parser.add_argument("-n", dest="max_iter", default=10000, type=int)
    parser.add_argument("-d", dest="min_improv", default=1e-10, type=float)
    # parser.add_argument(
    #     "-o",
    #     "--output",
    #     dest="output",
    #     type=str,
    #     default=None,
    #     help="write resulting model to given file",
    # )
    parser.add_argument(
        "-O",
        "--output-model",
        type=str,
        default=None,
        help="write resulting model to given file",
    )
    parser.add_argument(
        "-o", "--output", type=str, default=None, help="write output to file"
    )
    parser.add_argument(
        "--logger", type=str, default=None, help="write log to a given file"
    )
    parser.add_argument(
        "-k",
        "--knowledge",
        dest="koption",
        choices={"sdd", "sddx", "ddnnf"},  # get_evaluatables(),
        default=None,
        help="knowledge compilation tool",
    )
    parser.add_argument(
        "--logspace",
        action="store_true",
        help="use log space evaluation",
        default=False,
    )
    parser.add_argument(
        "-l",
        "--leak-probabilities",
        dest="leakprob",
        type=float,
        help="Add leak probabilities for evidence atoms",
    )
    parser.add_argument(
        "--propagate-evidence",
        action="store_true",
        dest="propagate_evidence",
        default=True,
        help="Enable evidence propagation",
    )
    parser.add_argument(
        "--dont-propagate-evidence",
        action="store_false",
        dest="propagate_evidence",
        default=True,
        help="Disable evidence propagation",
    )
    normalize_group = parser.add_mutually_exclusive_group()
    normalize_group.add_argument(
        "--normalize",
        action="store_true",
        dest="normalize",
        default=True,
        help="Normalize AD-weights (default).",
    )
    normalize_group.add_argument(
        "--nonormalize",
        action="store_false",
        dest="normalize",
        help="Do not normalize AD-weights.",
    )
    parser.add_argument("-v", "--verbose", action="count", default=0)
    parser.add_argument("--web", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument(
        "-a",
        "--arg",
        dest="args",
        action="append",
        help="Pass additional arguments to the cmd_args builtin.",
    )

    return parser


def create_logger(name, verbose):
    levels = [logging.WARNING, logging.INFO, logging.DEBUG] + list(range(9, 0, -1))
    verbose = max(0, min(len(levels) - 1, verbose))
    logger = getLogger(name)
    ch = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter("[%(levelname)s] %(message)s")
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(levels[verbose])


def lfi_wrapper(plfile, evfiles, knowledge, options):
    program = PrologFile(plfile)
    examples = list(read_examples(*evfiles))
    return run_lfi(program, examples, knowledge=get_evaluatable(knowledge), **options)


def main(argv, result_handler=None):
    parser = argparser()
    args = parser.parse_args(argv)

    if result_handler is None:
        if args.web:
            result_handler = print_result_json
        else:
            result_handler = print_result

    knowledge = get_evaluatable(args.koption)

    if args.output is None:
        outf = None
    else:
        outf = open(args.output, "w")

    if args.logger is None:
        logf = None
    else:
        logf = open(args.logger, "w")


    logger = init_logger(verbose=args.verbose, name="problog_lfi", out=logf)

    program = PrologFile(args.model)
    examples = list(read_examples(*args.examples))
    if len(examples) == 0:
        getLogger("problog_lfi").warning("no examples specified")
    else:
        getLogger("problog_lfi").info("Number of examples: %s" % len(examples))
    options = vars(args)
    del options["examples"]

    try:
        if args.seed is not None:
            random.seed(a=args.seed)
        results = run_lfi(program, examples, knowledge=knowledge, **options)

        for n in results[2]:
            n.loc = program.lineno(n.location)
        retcode = result_handler((True, results), outf=outf)
    except Exception as err:
        trace = traceback.format_exc()
        getLogger("problog_lfi").error("\nError encountered:\t\n" + trace)
        err.trace = trace
        retcode = result_handler((False, err), outf=outf)

    if args.logger is not None:
        logf.close()

    if args.output is not None:
        outf.close()

    if retcode:
        sys.exit(retcode)


def print_result(d, outf, precision=8):
    success, d = d
    if success:
        score, weights, names, iterations, lfi = d
        weights = list(map(lambda x: round(x, precision), weights))
        # print("Score: " + str(score), file=outf)
        # print("Tunable clauses: " + str(names), file=outf)
        # print("Learned weights: " + str(weights), file=outf)
        # print(f
        #     "Number of iterations before convergence: {iterations}", file=outf
        # )
        print(score, weights, names, iterations, file=outf)
        return 0
    else:
        print(process_error(d), file=outf)
        return 1


def print_result_json(d, outf, precision=8):
    import json

    success, d = d
    if success:
        score, weights, names, iterations, lfi = d
        results = {
            "SUCCESS": True,
            "score": score,
            "iterations": iterations,
            "weights": [
                [str(n), round(w, precision), n.loc[1], n.loc[2]]
                for n, w in zip(names, weights)
            ],
            "model": lfi.get_model(),
        }
        print(json.dumps(results), file=outf)
    else:
        results = {"SUCCESS": False, "err": vars(d)}
        print(json.dumps(results), file=outf)
    return 0


if __name__ == "__main__":
    main(sys.argv[1:])



import sys

from problog.cycles import break_cycles
from problog.engine import DefaultEngine
from problog.formula import LogicDAG
from problog.logic import Constant, Term
from problog.program import PrologFile, ExtendedPrologFactory
from problog.sdd_formula import SDD


class ConstraintFactory(ExtendedPrologFactory):
    """Extends standard ProbLog input reader to process ensure_true, ensure_false and ensure_prob."""

    def __init__(self, *args, **kwargs):
        super(ConstraintFactory, self).__init__(*args, **kwargs)
        self.constraint_count = 0

    def build_clause(self, functor, operand1, operand2, location=None, **extra):
        if len(operand1) == 1 and operand1[0].functor in (
            "ensure_true",
            "ensure_false",
            "ensure_prob",
        ):
            operand1 = [
                Term("constraint", Constant(self.constraint_count), operand1[0])
            ]
            self.constraint_count += 1
        return super(ConstraintFactory, self).build_clause(
            functor, operand1, operand2, location=location, **extra
        )


def formula_to_flatzinc_float(formula):
    """Converts a formula to a MiniZinc model using floats."""
    # print (formula)
    vardefs = []
    constraints = []
    boolvars = []

    for i, n, t in formula:
        if t == "atom":
            if str(n.probability) == "?":
                vardefs.append("var bool: B%s :: output_var;" % i)
                vardefs.append("var int: I%s;" % i)
                vardefs.append("var 0.0 .. 1.0: F%s;" % i)
                constraints.append("constraint bool2int(B%s,I%s);" % (i, i))
                constraints.append("constraint int2float(I%s,F%s);" % (i, i))
                boolvars.append("B%s" % i)
                vardefs.append("var 0.0 .. 1.0: V%s;" % i)
                vardefs.append("var 0.0 .. 1.0: Vn%s;" % i)
                constraints.append(
                    "constraint float_lin_eq([1.0, 1.0], [V%s, Vn%s], 1.0);" % (i, i)
                )
                constraints.append("constraint float_eq(F%s,V%s);" % (i, i))
            else:
                vardefs.append("float: V%s = %s;" % (i, n.probability))
                vardefs.append("float: Vn%s = %s;" % (i, 1 - float(n.probability)))
        elif t == "disj":
            vardefs.append("var 0.0 .. 1.0: V%s;" % i)
            vardefs.append("var 0.0 .. 1.0: Vn%s;" % i)
            constraints.append(
                "constraint float_lin_eq([1.0, 1.0], [V%s, Vn%s], 1.0);" % (i, i)
            )

            children = ["V%s" % str(c).replace("-", "n") for c in n.children] + [
                "V%s" % i
            ]
            coef = ["1.0"] * (len(children) - 1) + ["-1.0"]
            constraints.append(
                "constraint float_lin_eq([%s], [%s], 0.0);"
                % (", ".join(coef), ", ".join(children))
            )
        elif t == "conj":
            vardefs.append("var 0.0 .. 1.0: V%s;" % i)
            vardefs.append("var 0.0 .. 1.0: Vn%s;" % i)
            constraints.append(
                "constraint float_lin_eq([1.0, 1.0], [V%s, Vn%s], 1.0);" % (i, i)
            )

            children = [str(c).replace("-", "n") for c in n.children]
            constraints.append(
                "constraint float_times(V%s, V%s, V%s);" % (children[0], children[1], i)
            )

    for name, index in formula.get_names(label="constraint"):
        index = str(index).replace("-", "n")
        con = name.args[1]
        if con.functor == "ensure_prob":
            constraints.append(
                "constraint float_lin_le([1.0],[V%s], %s);" % (index, con.args[1])
            )
            constraints.append(
                "constraint float_lin_le([-1.0], [V%s], -%s);" % (index, con.args[0])
            )
        elif con.functor == "ensure_true":
            constraints.append("constraint float_eq(V%s, 1.0);" % (index,))
        elif con.functor == "ensure_false":
            constraints.append("constraint float_eq(V%s, 0.0);" % (index,))

    solve = [
        "solve :: bool_search([%s], input_order, indomain_min, complete) satisfy;"
        % ", ".join(boolvars)
    ]

    fzn = "\n".join(vardefs + constraints + solve)

    return fzn


def formula_to_flatzinc_bool(formula):
    """Converts a formula to a MiniZinc model using booleans."""
    vardefs = []
    constraints = []
    boolvars = []

    for i, n, t in formula:
        if t == "atom":
            if str(n.probability) == "?":
                vardefs.append("var bool: B%s :: output_var;" % i)
            else:
                vardefs.append("bool: B%s = true;" % (i,))
        elif t == "disj":
            vardefs.append("var bool: B%s;" % i)
            children = ["B%s" % str(c).replace("-", "n") for c in n.children]
            constraints.append(
                "constraint array_bool_or([%s], %s);" % (", ".join(children), "B%s" % i)
            )
        elif t == "conj":
            vardefs.append("var bool: B%s;" % i)
            children = ["B%s" % str(c).replace("-", "n") for c in n.children]
            constraints.append(
                "constraint array_bool_and([%s], %s);"
                % (", ".join(children), "B%s" % i)
            )
        vardefs.append("var bool: Bn%s;" % i)
        constraints.append("constraint bool_not(B%s, Bn%s);" % (i, i))

    for name, index in formula.get_names(label="constraint"):
        if index is not None:
            index = str(index).replace("-", "n")
            con = name.args[1]
            assert con.functor != "ensure_prob"
            if con.functor == "ensure_true":
                constraints.append("constraint bool_eq(B%s, true);" % (index,))
            elif con.functor == "ensure_false":
                constraints.append("constraint bool_eq(B%s, false);" % (index,))

    solve = [
        "solve :: bool_search([%s], input_order, indomain_min, complete) satisfy;"
        % ", ".join(boolvars)
    ]

    fzn = "\n".join(vardefs + constraints + solve)

    return fzn


def debug(*args):
    """Helper function for printing debugging output."""
    print(*args, file=sys.stderr)


def find_solver():
    import subprocess

    try:
        subprocess.check_output(["fzn-gecode", "-help"])
    except FileNotFoundError:
        raise ImportError(
            "Please install 'fzn-gecode' to use the constraint task: 'apt install flatzinc'."
        )


def solve(fzn):
    """Call FlatZinc solver (Gecode) and process output."""
    import subprocess

    from problog.util import mktempfile

    fznfile = mktempfile(".fzn")
    with open(fznfile, "w") as f:
        f.write(fzn)
    cmd = ["fzn-gecode", "-a", fznfile]
    result = subprocess.check_output(cmd).decode()

    current = {}
    for line in result.split("\n"):
        if line.startswith("----"):
            yield current
            current = {}
        elif line.startswith("===="):
            return
        else:
            vr, vl = line.strip(";").split(" = ")
            vr = int(vr[1:])
            vl = 1 if vl == "true" else 0
            current[vr] = vl


def compress(formula, atoms):
    """Compress the formula by setting the values of the given atoms."""

    # TODO support cyclic programs

    formula.clear_labeled("constraint")

    out = LogicDAG()

    relevant = formula.extract_relevant()

    cyclic = set()

    translate = {}
    for i, n, t in formula:

        if not relevant[i]:
            continue
        elif t == "atom":
            if i in atoms:
                if atoms[i] == 1:
                    translate[i] = 0
                else:
                    translate[i] = None
            else:
                translate[i] = out.add_atom(*n)
        else:
            children = []
            for c in n.children:
                if abs(c) in translate:
                    tr = translate[abs(c)]
                    if c > 0:
                        children.append(tr)
                    else:
                        children.append(out.negate(tr))
                else:
                    cyclic.add(c)
            if t == "conj":
                translate[i] = out.add_and(children)
            else:
                translate[i] = out.add_or(children)

    if cyclic:
        debug("Original formula:\n", formula)
        debug("Cyclic nodes:", cyclic)
        debug("New formula so far:\n", out)
        raise RuntimeError(
            "Cycle detected in program. Cycles are currently not supported."
        )

    for q, n, l in formula.labeled():
        if l != "constraint":
            if n < 0:
                out.add_name(q, out.negate(translate[-n]), l)
            else:
                out.add_name(q, translate[n], l)
    return out


def create_sdd(solution, formula, verbose):
    if verbose:
        debug("Compressing...")
    new_formula = compress(formula, solution)

    if verbose:
        debug("Compiling...")
    current_sdd = SDD.create_from(new_formula)

    if verbose:
        debug("Evaluating...")
    return current_sdd


def enumerate_sdd(solution, formula, sdd, verbose):
    # target.lookup_evidence = {}
    weights = sdd.get_weights()
    # ev_nodes = []
    for k, v in solution.items():
        sddvar = formula.get_node(k).identifier
        sddatom = sdd.var2atom[sddvar]
        sddname = sdd.get_name(sddatom)
        weights[sddatom] = v
    return sdd


def enumerate_solutions(sols, verbose, formula, sdd=None):
    solutions = []
    for i, res in enumerate(sols):
        if verbose:
            debug("Evaluating solution %s/%s..." % (i + 1, len(sols)))

        if sdd is not None:
            current_sdd = enumerate_sdd(res, formula, sdd, verbose)
        else:
            current_sdd = create_sdd(res, formula, verbose)

        solutions.append(current_sdd.evaluate().items())
    return solutions


def check_prob_constraint(constraints):
    return any(
        index is not None and name.args[1].functor == "ensure_prob"
        for name, index in constraints
    )


def main(argv, handle_output=None):
    args = argparser().parse_args(argv)

    verbose = args.verbose
    filename = args.filename
    solutions = run(filename, verbose)

    for solution in solutions:
        for k, v in solution:
            if v > 0.0:
                print(k, v)

        print("----------")

    if solutions:
        print("==========")
    else:
        print("=== UNSATISFIABLE ===")

    if verbose:
        debug("Loading...")


def run(filename, verbose=None):
    find_solver()
    if verbose:
        debug("Loading...")
    problog_model = PrologFile(filename, factory=ConstraintFactory())

    engine = DefaultEngine()
    database = engine.prepare(problog_model)

    # Ground the constraints
    if verbose:
        debug("Grounding...")
    target = engine.ground(database, Term("constraint", None, None), label="constraint")

    queries = [q[0] for q in engine.query(database, Term("query", None))]
    for query in queries:
        target = engine.ground(database, query, label="query", target=target)

    if verbose is not None and verbose > 1:
        print(target, file=sys.stderr)

    has_prob_constraint = check_prob_constraint(target.get_names(label="constraint"))

    # Compile and turn into CP-problem
    sdd = None
    if has_prob_constraint:
        if verbose:
            debug("Probabilistic constraints detected.")
            debug("Compiling...")
        sdd = SDD.create_from(target)
        formula = sdd.to_formula()
    else:
        formula = LogicDAG()
        break_cycles(target, formula)

    if verbose:
        debug("Converting...")
    if sdd:
        fzn = formula_to_flatzinc_float(formula)
        sdd.clear_labeled("constraint")
    else:
        fzn = formula_to_flatzinc_bool(formula)

    if verbose is not None and verbose > 1:
        print(fzn, file=sys.stderr)
    if verbose:
        debug("Solving...")
    sols = list(solve(fzn))

    return enumerate_solutions(sols, verbose, formula, sdd)


def argparser():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("filename", metavar="MODEL")
    parser.add_argument("-v", "--verbose", action="count")
    return parser


if __name__ == "__main__":
    main(sys.argv[1:])



"""
ProbLog IPython magic extensions

Magic methods:
    %problog <model>
    %%problog <model ...
    ... >
    %problogstr "<dot graph>"
    %problogmodel model
    %problogmodels [model1, model2, ...]

    %%problogsample [-N int]
    <model>

Usage:

    %load_ext problogmagic

Based on:
https://gist.github.com/cjdrake/7982333
"""

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(os.path.dirname(__file__), "../examples"))

from problog.logic import Term
from problog.program import PrologString, ExtendedPrologFactory, LogicProgram
from problog.evaluator import SemiringLogProbability
from problog.parser import DefaultPrologParser
from problog.ddnnf_formula import DDNNF
from problog.sdd_formula import SDD
import problog.tasks.sample as plsample

if SDD.is_available():
    knowledge_default = SDD
else:
    knowledge_default = DDNNF

from IPython.core.display import display_html
from IPython.core.magic import Magics, magics_class, line_magic, line_cell_magic
from IPython.core.magic_arguments import argument, magic_arguments, parse_argstring

import IPython

if IPython.version_info[0] >= 6:
    import warnings

    def info(message):
        warnings.warn("WARNING: " + message)

    def error(message):
        warnings.warn("ERROR: " + message)


else:
    pass


def runproblog(
    s,
    knowledge=knowledge_default,
    semiring=None,
    parser_class=DefaultPrologParser,
    output="html",
):
    """Execute problog and return an html snippet, or None."""
    try:
        parser = parser_class(ExtendedPrologFactory())
        model = PrologString(s, parser=parser)
        formula = knowledge.createFrom(model)
        result = formula.evaluate(semiring=semiring)
        # result = ProbLog.convert(model, knowledge).evaluate()
    except Exception as e:
        return "<pre>{}</pre>".format(e)
    if result is None:
        warnings.warn("Error when running ProbLog")
        return None
    else:
        return formatoutput(result, output=output)


def runproblogsampling(s, n=5, output="html"):
    model = PrologString(s)
    samples = plsample.sample(model, n=n, tuples=True)
    result = ""
    for sample in samples:
        result += (
            ",".join(str(Term(query[0], *query[1:-1])) for query in sample) + "<br/>"
        )
    if output == "html":
        return "<pre>{}</pre>".format(result)
    return result


def formatoutput(result, output="html"):
    if output == "html":
        html = (
            '<table style="width:100%;"><tr><th style="width:66%;">Atom<th>Probability'
        )
        atomprobs = [(str(atom), prob) for atom, prob in result.items()]
        atomprobs.sort(key=lambda x: x[0])
        for atom, prob in atomprobs:
            p = prob * 100
            html += "<tr><td>{a}".format(a=atom)
            color = "black"
            html += (
                '<td><div class="progress-bar" role="progressbar" aria-valuenow="{perc}" '
                'aria-valuemin="0" aria-valuemax="100" style="text-align:left;'
                'width:{perc}%;padding:3px;color:{c};background-color:#9ac2f4;">'
                "{prob:6.4f}</div>".format(perc=prob * 100, prob=prob, c=color)
            )
        html += "</table>"
        return html
    else:
        txt = ""
        for atom, prob in result.items():
            txt += "{:<40}: {}".format(atom, prob)
        return txt


@magics_class
class ProbLogMagics(Magics):
    @line_cell_magic
    @magic_arguments()
    @argument(
        "--knowledge",
        "-k",
        choices=("sdd", "nnf"),
        default=None,
        help="Knowledge compilation tool.",
    )
    @argument("--logspace", action="store_true", help="Use log space evaluation.")
    def problog(self, line, cell=None):
        """problog line/cell magic"""
        args = parse_argstring(self.problog, line)
        if args.knowledge == "nnf":
            knowledge = DDNNF
        elif args.knowledge == "sdd":
            knowledge = SDD
        elif args.knowledge is None:
            if SDD.is_available():
                knowledge = SDD
            else:
                knowledge = DDNNF
        else:
            warnings.warn("Unknown option for --knowledge: '%s'" % args.knowledge)
        if args.logspace:
            semiring = SemiringLogProbability()
        else:
            semiring = None

        s = cell
        data = runproblog(s, knowledge=knowledge, semiring=semiring, output="html")
        if data:
            display_html(data, raw=True)

    @line_magic
    def problogstr(self, line):
        """problog string magic"""
        s = self.shell.ev(line)
        data = runproblog(s, output="html")
        if data:
            display_html(data, raw=True)

    @line_magic
    def problogobj(self, line):
        """problog object magic"""
        obj = self.shell.ev(line)
        if not isinstance(obj, LogicProgram):
            warnings.warn("expected object to be of type LogicProgram")
        else:
            data = runproblog(obj, output="html")
            if data:
                display_html(data, raw=True)

    @line_magic
    def problogobjs(self, line):
        """problog objects magic"""
        objs = self.shell.ev(line)
        for i, obj in enumerate(objs):
            if not isinstance(obj, LogicProgram):
                warnings.warn("expected object {} to be of type LogicProgram".format(i))
            else:
                data = runproblog(obj, output="html")
                if data:
                    warnings.warn("object {}:".format(i))
                    display_html(data, raw=True)

    @line_cell_magic
    @magic_arguments()
    @argument("-N", help="Number of samples")
    def problogsample(self, line, cell=None):
        """problog line/cell magic"""
        args = parse_argstring(self.problogsample, line)
        n = 5
        if args.N is not None:
            n = int(args.N)
        s = cell
        data = runproblogsampling(s, n=n, output="html")
        if data:
            display_html(data, raw=True)


def load_ipython_extension(ipython):
    """Load the extension in IPython."""
    ipython.register_magics(ProbLogMagics)



import os
import unittest
from pathlib import Path

from problog.logic import Term, Constant
from problog.tasks import dtproblog

dirname = os.path.dirname(__file__)
dt_problog_test_folder = Path(dirname, "./../../test/dtproblog/")


class TestDTProblog(unittest.TestCase):
    def dt_problog_check_if_output_equals(
        self, dt_file, expected_choices, expected_score, local=False
    ):
        real_file_name = dt_problog_test_folder / dt_file
        file_exists = real_file_name.exists()
        if not file_exists:
            self.fail(
                "File "
                + str(real_file_name)
                + " was not found. Maybe this is a pathing issue?"
            )
        else:
            arguments = [str(real_file_name)]
            if local:
                arguments.append("-s")
                arguments.append("local")

            result = dtproblog.main(arguments)
            if result[0]:
                choices, score, stats = result[1]
                self.assertEqual(expected_choices, choices)
                self.assertAlmostEqual(expected_score, score, delta=1e-6)
            else:
                self.fail(
                    "An error occured was returned when executing "
                    + dt_file
                    + "\nError "
                    + str(result[1])
                )

    def test_dt_problog_simple(self):
        self.dt_problog_check_if_output_equals("ex1.pl", {Term("b"): 1}, 1.8)
        self.dt_problog_check_if_output_equals("ex2.pl", {Term("c"): 1}, 1.8)
        self.dt_problog_check_if_output_equals("ex3.pl", {Term("a"): 1}, 1.8)
        self.dt_problog_check_if_output_equals(
            "ex4.pl", {Term("b"): 1, Term("c"): 1}, 3
        )
        self.dt_problog_check_if_output_equals(
            "ex5.pl", {Term("a"): 1, Term("c"): 1}, 1.8
        )

    def test_dt_problog_local(self):
        self.dt_problog_check_if_output_equals(
            "ex5.pl", {Term("a"): 0, Term("c"): 0}, 0, local=True
        )
        # Takes too long:
        # self.dt_problog_check_if_output_equals(
        #     "umbrella_poole.pl",
        #     {
        #         Term("umbrella", Term("sunny")): 0,
        #         Term("umbrella", Term("cloudy")): 0,
        #         Term("umbrella", Term("rainy")): 1,
        #     },
        #     77, local=True
        # )

    def test_dtproblog_mutex(self):
        self.dt_problog_check_if_output_equals(
            "mut_exl.pl",
            {
                Term("play", Constant(1)): 0,
                Term("play", Constant(2)): 1,
                Term("bet", Constant(1), Term("heads")): 0,
                Term("bet", Constant(2), Term("heads")): 0,
                Term("bet", Constant(1), Term("tails")): 1,
                Term("bet", Constant(2), Term("tails")): 1,
            },
            1.5,
        )
        self.dt_problog_check_if_output_equals(
            "mut_exl2.pl",
            {
                Term("play"): 0,
                Term("bet", Term("heads")): 0,
                Term("bet", Term("tails")): 1,
            },
            0,
        )

    def test_dtproblog_umbrella(self):
        self.dt_problog_check_if_output_equals(
            "umbrella.pl", {Term("umbrella"): 1, Term("raincoat"): 0}, 43
        )
        self.dt_problog_check_if_output_equals(
            "umbrella_poole.pl",
            {
                Term("umbrella", Term("sunny")): 0,
                Term("umbrella", Term("cloudy")): 0,
                Term("umbrella", Term("rainy")): 1,
            },
            77,
        )

    # Takes too long:

    # def test_dtproblog_long_test(self):
    #     self.dt_problog_check_if_output_equals(
    #         "viralmarketing.pl",
    #         {
    #             Term("marketed", Term("angelika")): 0,
    #             Term("marketed", Term("bernd")): 0,
    #             Term("marketed", Term("guy")): 1,
    #             Term("marketed", Term("ingo")): 1,
    #             Term("marketed", Term("kurt")): 0,
    #             Term("marketed", Term("laura")): 0,
    #             Term("marketed", Term("martijn")): 1,
    #             Term("marketed", Term("theo")): 1,
    #         },
    #         3.210966333135799,
    #         local=True
    #     )

    def test_dtproblog_other(self):
        self.dt_problog_check_if_output_equals(
            "var_util.pl",
            {
                Term("bet", Term("w1"), Constant(5)): 1,
                Term("bet", Term("w2"), Constant(7)): 0,
                Term("bet", Term("w3"), Constant(10)): 0,
            },
            4,
        )
        self.dt_problog_check_if_output_equals(
            "weather_poole_alt.pl",
            {
                Term("decide_u", Term("rainy")): 1,
                Term("decide_u", Term("cloudy")): 0,
                Term("decide_u", Term("sunny")): 0,
            },
            77,
        )
        # self.dt_problog_check_if_output_equals("winning.pl", {}, 0)
        self.dt_problog_check_if_output_equals(
            "winning_undecided.pl", {Term("play1"): 1, Term("play2"): 1}, 34.5
        )

    if __name__ == "__main__":
        unittest.main()



"""
problog.cycles - Cycle-breaking
-------------------------------

Cycle breaking in propositional formulae.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import logging
from collections import defaultdict

from .core import transform
from .formula import LogicFormula, LogicDAG
from .logic import Term
from .util import Timer

cycle_var_prefix = "problog_cv_"

# noinspection PyUnusedLocal
@transform(LogicFormula, LogicDAG)
def break_cycles(source, target, translation=None, keep_named=False, **kwdargs):
    """Break cycles in the source logic formula.

    :param source: logic formula with cycles
    :param target: target logic formula without cycles
    :param keep_named: if true, then named nodes will be preserved after cycle breaking
    :param kwdargs: additional arguments (ignored)
    :return: target
    """

    logger = logging.getLogger("problog")
    with Timer("Cycle breaking"):
        cycles_broken = set()
        content = set()
        if translation is None:
            translation = defaultdict(list)

        labeled = source.labeled()
        if keep_named:
            for name, node, label in source.get_names_with_label():
                if label == source.LABEL_NAMED:
                    labeled.append((name, node, label))
        for q, n, l in labeled:
            if source.is_probabilistic(n):
                newnode = _break_cycles(
                    source, target, n, [], cycles_broken, content, translation
                )
            else:
                newnode = n
            target.add_name(q, newnode, l)

        # TODO copy constraints

        translation = defaultdict(list)
        for q, n, v in source.evidence_all():
            if source.is_probabilistic(n):
                newnode = _break_cycles(
                    source,
                    target,
                    abs(n),
                    [],
                    cycles_broken,
                    content,
                    translation,
                    is_evidence=True,
                )
            else:
                newnode = n
            if n is not None and n < 0:
                newnode = target.negate(newnode)
            if v > 0:
                target.add_name(q, newnode, target.LABEL_EVIDENCE_POS)
            elif v < 0:
                target.add_name(q, newnode, target.LABEL_EVIDENCE_NEG)
            else:
                target.add_name(q, newnode, target.LABEL_EVIDENCE_MAYBE)

        logger.debug("Ground program size: %s", len(target))
        return target


def _break_cycles(
    source,
    target,
    nodeid,
    ancestors,
    cycles_broken,
    content,
    translation,
    is_evidence=False,
):
    negative_node = nodeid < 0
    nodeid = abs(nodeid)

    if not is_evidence and not source.is_probabilistic(
        source.get_evidence_value(nodeid)
    ):
        ev = source.get_evidence_value(nodeid)
        if negative_node:
            return target.negate(ev)
        else:
            return ev
    elif nodeid in ancestors:
        cycles_broken.add(nodeid)
        return None  # cyclic node: node is False
    elif nodeid in translation:
        ancset = frozenset(ancestors + [nodeid])
        for newnode, cb, cn in translation[nodeid]:
            # We can reuse this previous node iff
            #   - no more cycles have been broken that should not be broken now
            #       (cycles broken is a subset of ancestors)
            #   - no more cycles should be broken than those that have been broken in the previous
            #       (the previous node does not contain ancestors)

            if cb <= ancset and not ancset & cn:
                cycles_broken |= cb
                content |= cn
                if negative_node:
                    return target.negate(newnode)
                else:
                    return newnode

    child_cycles_broken = set()
    child_content = set()

    node = source.get_node(nodeid)
    nodetype = type(node).__name__
    if nodetype == "atom":
        newnode = target.add_atom(
            node.identifier,
            node.probability,
            group=node.group,
            name=node.name,
            is_extra=node.is_extra,
        )
    else:
        children = [
            _break_cycles(
                source,
                target,
                child,
                ancestors + [nodeid],
                child_cycles_broken,
                child_content,
                translation,
                is_evidence,
            )
            for child in node.children
        ]
        newname = node.name
        if newname is not None and child_cycles_broken:
            newfunc = (
                cycle_var_prefix + newname.functor + "_cb_" + str(len(translation[nodeid]))
            )
            newname = Term(newfunc, *newname.args)
        if nodetype == "conj":
            newnode = target.add_and(children, name=newname)
        else:
            newnode = target.add_or(children, name=newname)

        if target.is_probabilistic(newnode):
            # Don't add the node if it is None
            # Also: don't add atoms (they can't be involved in cycles)
            content.add(nodeid)

    translation[nodeid].append(
        (newnode, child_cycles_broken, child_content - child_cycles_broken)
    )
    content |= child_content
    cycles_broken |= child_cycles_broken

    if negative_node:
        return target.negate(newnode)
    else:
        return newnode



"""
problog.constraint - Propositional constraints
----------------------------------------------

Data structures for specifying propositional constraints.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from .errors import InvalidValue
from .logic import Term, Constant


class Constraint(object):
    """A propositional constraint."""

    def get_nodes(self):
        """Get all nodes involved in this constraint."""
        raise NotImplementedError("abstract method")

    def update_weights(self, weights, semiring):
        """Update the weights in the given dictionary according to the constraints.

        :param weights: dictionary of weights (see result of :func:`LogicFormula.extract_weights`)
        :param semiring: semiring to use for weight transformation
        """
        # Typically, constraints don't update weights
        pass

    def is_true(self):
        """Checks whether the constraint is trivially true."""
        return False

    def is_false(self):
        """Checks whether the constraint is trivially false."""
        return False

    def is_nontrivial(self):
        """Checks whether the constraint is non-trivial."""
        return not self.is_true() and not self.is_false()

    def as_clauses(self):
        """Represent the constraint as a list of clauses (CNF form).

        :return: list of clauses where each clause is represent as a list of node keys
        :rtype: list[list[int]]
        """
        raise NotImplementedError("abstract method")

    def copy(self, rename=None):
        """Copy this constraint while applying the given node renaming.

        :param rename: node rename map (or None if no rename is required)
        :return: copy of the current constraint
        """
        raise NotImplementedError("abstract method")


class ConstraintAD(Constraint):
    """Annotated disjunction constraint (mutually exclusive with weight update)."""

    def __init__(self, group):
        self.nodes = set()
        self.group = group
        self.extra_node = None
        self.location = None

    def __str__(self):
        return "annotated_disjunction(%s, %s)" % (list(self.nodes), self.extra_node)

    def get_nodes(self):
        if self.extra_node:
            return list(self.nodes) + [self.extra_node]
        else:
            return self.nodes

    def is_true(self):
        return len(self.nodes) <= 1

    def is_false(self):
        return False

    def add(self, node, formula, cr_extra=True):
        """Add a node to the constraint from the given formula.

        :param node: node to add
        :param formula: formula from which the node is taken
        :param cr_extra: Create an extra_node when required (when it is None and this is the second atom of the group).
        :return: value of the node after constraint propagation
        """

        if node in self.nodes:
            return node

        is_extra = formula.get_node(node).is_extra

        try:
            if (
                not self.location
                and formula.get_node(node).name
                and formula.get_node(node).name.args
            ):
                if formula.database:
                    self.location = formula.database.lineno(
                        formula.get_node(node).name.args[-1].location
                    )
        except AttributeError:
            pass

        if formula.has_evidence_values() and not is_extra:
            # Propagate constraint: if one of the other nodes is True: this one is false
            for n in self.nodes:
                if formula.get_evidence_value(n) == formula.TRUE:
                    return formula.FALSE
            if formula.get_evidence_value(node) == formula.FALSE:
                return node
            elif formula.get_evidence_value(node) == formula.TRUE:
                for n in self.nodes:
                    formula.set_evidence_value(n, formula.FALSE)

            if formula.semiring:
                sr = formula.semiring
                w = formula.get_weight(node, sr)
                for n in self.nodes:
                    w = sr.plus(w, formula.get_weight(n, sr))
                if sr.is_one(w):
                    unknown = None
                    if formula.get_evidence_value(node) != formula.FALSE:
                        unknown = node
                    for n in self.nodes:
                        if formula.get_evidence_value(n) != formula.FALSE:
                            if unknown is not None:
                                unknown = None
                                break
                            else:
                                unknown = n
                    if unknown is not None:
                        formula.set_evidence_value(unknown, formula.TRUE)

        if is_extra:
            self.extra_node = node
        else:
            self.nodes.add(node)

        if cr_extra and len(self.nodes) > 1 and self.extra_node is None:
            # If there are two or more choices -> add extra choice node
            self._update_logic(formula)
        return node

    def as_clauses(self):
        if self.is_nontrivial():
            nodes = list(self.nodes) + [self.extra_node]
            lines = []
            for i, n in enumerate(nodes):
                for m in nodes[i + 1 :]:
                    lines.append((-n, -m))  # mutually exclusive
            lines.append(nodes)  # pick one
            return lines
        else:
            return []

    def _update_logic(self, formula):
        """Add extra information to the logic structure of the formula.

        :param formula: formula to update
        """
        if self.is_nontrivial():
            name = Term(
                "choice",
                Constant(self.group[0]),
                Term("e"),
                Term("null"),
                *self.group[1]
            )
            self.extra_node = formula.add_atom(
                ("%s_extra" % (self.group,)),
                True,
                name=name,
                group=self.group,
                is_extra=True,
            )
            # formula.addConstraintOnNode(self, self.extra_node)

    def update_weights(self, weights, semiring):
        if self.is_nontrivial():
            ws = []
            for n in self.nodes:
                pos, neg = weights.get(n, (semiring.one(), semiring.one()))
                weights[n] = (pos, semiring.ad_negate(pos, neg))
                ws.append(pos)

            name = Term(
                "choice",
                Constant(self.group[0]),
                Term("e"),
                Term("null"),
                *self.group[1]
            )
            try:
                complement = semiring.ad_complement(ws, key=name)
                if not semiring.in_domain(complement):
                    raise InvalidValue(
                        "Sum of annotated disjunction weigths exceeds acceptable value",
                        location=self.location,
                    )
            except InvalidValue:
                raise InvalidValue(
                    "Sum of annotated disjunction weigths exceeds acceptable value",
                    location=self.location,
                )
                # TODO add location
            weights[self.extra_node] = (
                complement,
                semiring.ad_negate(complement, semiring.one()),
            )

    def copy(self, rename=None):
        if rename is None:
            rename = {}
        result = ConstraintAD(self.group)
        result.nodes = set(rename.get(x, x) for x in self.nodes)
        result.extra_node = rename.get(self.extra_node, self.extra_node)
        return result

    def check(self, values):
        """Check the constraint

        :param values: dictionary of values for nodes
        :return: True if constraint succeeds, False otherwise
        """
        if self.is_true():
            return True
        elif self.is_false():
            return False
        else:
            actual = [
                values.get(i) for i in self.get_nodes() if values.get(i) is not None
            ]
            return sum(actual) == 1

    def propagate(self, values, weights, node=None):
        """Returns
            - True: constraint satisfied
            - False: constraint violated
            - None: unknown
        """

        if node is not None and node not in self.get_nodes():
            return

        if self.is_true():
            return True
        elif self.is_false():
            return False
        else:
            # print ([(i, values[i]) for i in self.get_nodes() if values.get(i) is not None], self.get_nodes())
            # If there is a true value: set all the others to false
            true_values = [i for i in self.get_nodes() if values.get(i) == 1.0]
            if len(true_values) == 1:
                v = true_values[0]
                for i in self.get_nodes():
                    if i != v:
                        values[i] = 0.0
                #     print ('a', values)
                return True
            elif len(true_values) > 1:
                #       print('b', values)
                return False
            else:
                false_values = set(
                    [i for i in self.get_nodes() if values.get(i) == 0.0]
                )
                remain = 1.0 - sum(weights[v] for v in false_values)
                # if len(false_values) == len(self.get_nodes()) - 1:
                for i in self.get_nodes():
                    if not i in false_values:
                        values[i] = weights[i] / remain
                #      print ('c', values)
                return True
        # print ('d', values)


class ClauseConstraint(Constraint):
    """A constraint specifying that a given clause should be true."""

    def __init__(self, nodes):
        self.nodes = nodes

    def as_clauses(self):
        return [self.nodes]

    def copy(self, rename=None):
        if rename is None:
            rename = {}
        return ClauseConstraint(map(lambda x: rename.get(x, x), self.nodes))

    def __str__(self):
        return "%s is true" % self.nodes


class TrueConstraint(Constraint):
    """A constraint specifying that a given node should be true."""

    def __init__(self, node):
        self.node = node

    def get_nodes(self):
        return [self.node]

    def as_clauses(self):
        return [[self.node]]

    def copy(self, rename=None):
        if rename is None:
            rename = {}
        return TrueConstraint(rename.get(self.node, self.node))

    def __str__(self):
        return "%s is true" % self.node



import argparse
import time

from .. import get_evaluatables, get_evaluatable
from ..engine import DefaultEngine
from ..formula import LogicFormula, LogicDAG
from ..program import SimpleProgram, PrologFile


def argparser():
    ap = argparse.ArgumentParser()

    # TODO support more options of task 'prob' (e.g. -k)

    ap.add_argument(
        "filename", nargs="+", help="Name of file to run (can be multiple)."
    )
    ap.add_argument(
        "-b", "--base", action="append", default=[], help="Additional models to load."
    )
    ap.add_argument(
        "-n", "--repeat", type=int, default=1, help="Repeat each run this many times."
    )
    ap.add_argument("-k", choices=get_evaluatables())

    return ap


def main(argv):
    parser = argparser()
    args = parser.parse_args(argv)

    first = True
    all_timers = dict()
    for filename in args.filename:
        timer_list = []
        for run in range(0, args.repeat):
            timers = process(filename, base=args.base, ktype=args.k)
            if first:
                print("filename", timers.header, "total", sep=";")
                first = False
            print(filename, timers, timers.total, sep=";")
            timer_list.append(timers)
        all_timers[filename] = timer_list
    return all_timers


class Timer(object):
    def __init__(self, name):
        self.name = name
        self._start_time = None
        self._end_time = None
        self._err_name = None

    def __enter__(self):
        self._start_time = time.time()

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self._end_time = None
            self._err_name = exc_type.__name__
        else:
            self._end_time = time.time()
        return

    @property
    def elapsed_time(self):
        if self._err_name is not None:
            return self._err_name
        else:
            return self._end_time - self._start_time

    @property
    def elapsed_time_string(self):
        if self._err_name is not None:
            return self._err_name
        else:
            return "%.6f" % (self._end_time - self._start_time)


class TimerCollection(object):
    def __init__(self):
        self.timers = []

    def new(self, name):
        tmr = Timer(name)
        self.timers.append(tmr)
        return tmr

    @property
    def header(self):
        return ";".join([tmr.name for tmr in self.timers])

    def __iter__(self):
        for t in self.timers:
            yield t

    def __str__(self):
        return ";".join([tmr.elapsed_time_string for tmr in self.timers])

    @property
    def total(self):
        try:
            return sum(tmr.elapsed_time for tmr in self.timers)
        except TypeError:
            return "ERROR"

    def __len__(self):
        return len(self.timers)


def process(filename, base=None, ktype=None):
    if base is None:
        base = []

    timers = TimerCollection()

    # Step 1: parse the model
    with timers.new("parse"):
        model = SimpleProgram()
        for i, fn in enumerate([filename] + base):
            filemodel = PrologFile(fn)
            for line in filemodel:
                model += line
            if i == 0:
                model.source_root = filemodel.source_root

    # Step 2: compile the model into a database
    with timers.new("load"):
        engine = DefaultEngine()
        database = engine.prepare(model)

    # Step 3: ground all
    with timers.new("ground"):
        ground_program = LogicFormula.create_from(database)

    # Step 4: break cycles
    with timers.new("cycles"):
        if ktype == "fsdd":
            ground_formula = ground_program
        else:
            ground_formula = LogicDAG.create_from(ground_program)

    # Step 5: compile
    with timers.new("compile"):
        knowledge = get_evaluatable(ktype).create_from(ground_formula)

    # Step 6: evaluate
    with timers.new("evaluate"):
        results = knowledge.evaluate()

    return timers



"""
Module name
"""

import os
import sys
import unittest

from problog import root_path
from problog.util import subprocess_call, subprocess_check_output


class TestInterfaces(unittest.TestCase):
    def _test_cmd(self, task, testfile=None):
        problogcli = root_path("problog-cli.py")
        if testfile is None:
            testfile = [root_path("test", "7_probabilistic_graph.pl")]

        with open(os.devnull, "w") as out:
            if task is None:
                self.assertEqual(
                    subprocess_call(
                        [sys.executable, problogcli] + testfile, stdout=out
                    ),
                    0,
                )
            else:
                self.assertEqual(
                    subprocess_call(
                        [sys.executable, problogcli, task] + testfile, stdout=out
                    ),
                    0,
                )

    def test_cli_mpe(self):
        return self._test_cmd("mpe")

    def test_cli_prob(self):
        return self._test_cmd("prob")

    # def test_cli_explain(self):
    #     return self._test_cmd('explain')

    def test_cli_ground(self):
        return self._test_cmd("ground")

    def test_cli_sample(self):
        return self._test_cmd("sample")

    def test_cli_default(self):
        return self._test_cmd(None)

    def test_cli_learn(self):
        problogcli = root_path("problog-cli.py")

        model = root_path("test", "lfi", "ad", "ADtest_8_1.pl")
        examples = root_path("test", "lfi", "ad", "ADtest_8_1.ev")

        out = subprocess_check_output(
            [sys.executable, problogcli, "lfi", model, examples, ]
        )
        assert "[0.4, 0.2, 0.4] [t(_)::a(X), t(_)::b(Y), t(_)::c(Z)]" in out



import unittest

from problog.logic import And, AnnotatedDisjunction, Or, Clause, Not, Term, Var


class TestLogic(unittest.TestCase):
    def test_and(self):
        c1 = And(Term("a"), Term("b"))
        c2 = Term("a") & Term("b")
        c3 = Term("b") & Term("c")
        c4 = Term("a") & Term("b", None)
        c5 = Term("a") & Term("b") & Term("c")
        c6 = And(Term("a"), And(Term("b"), Term("c")))
        c7 = And(And(Term("a"), Term("b")), Term("c"))
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 == c3)
        self.assertFalse(c1 == c4)
        self.assertFalse(c2 == c4)
        self.assertTrue(c5 == c6)
        self.assertFalse(c5 == c7)
        self.assertFalse(c6 == c7)

    def test_or(self):
        c1 = Or(Term("a"), Term("b"))
        c2 = Term("a") | Term("b")
        c3 = Term("b") | Term("c")
        c4 = Term("a") | Term("b", None)
        c5 = Term("a") | Term("b") | Term("c")
        c6 = Or(Term("a"), Or(Term("b"), Term("c")))
        c7 = Or(Or(Term("a"), Term("b")), Term("c"))
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 == c3)
        self.assertFalse(c1 == c4)
        self.assertFalse(c2 == c4)
        self.assertTrue(c5 == c6)
        self.assertFalse(c5 == c7)
        self.assertFalse(c6 == c7)

    def test_and_or(self):
        c1 = Term("a") & Term("b") | Term("c")
        c2 = Term.from_string("(a,b);c.")
        c3 = Term("a") & (Term("b") | Term("c"))
        c4 = (Term("b") | Term("c")) & Term("a")
        self.assertTrue(c1 == c2)
        self.assertFalse(c2 == c3)
        self.assertFalse(c1 == c3)
        self.assertFalse(c3 == c4)  # Different because of unification

    def test_not(self):
        c1 = Not("\\+", Term("a"))
        c2 = ~Term("a")
        c3 = Not("not", Term("a"))
        self.assertTrue(c1 == c2)
        self.assertTrue(c1 == c3)

        c4 = And(Term("a"), Not("not", Term("b")))
        c5 = And(Term("a"), Not("\\+", Term("b")))
        self.assertTrue(c4 == c5)

    def test_lshift(self):
        c1 = Clause(Term("a"), Term("b"))
        c2 = Term("a") << Term("b")
        c3 = Term("b") << Term("c")
        c4 = Term("a") << Term("b", None)
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 == c3)
        self.assertFalse(c1 == c4)
        self.assertFalse(c2 == c4)

    def test_clause_equality(self):
        c1 = Clause(Term("a", None), Term("b", None))
        c2 = Clause(Term("a", None), Term("b", None))
        c3 = Clause(Term("a", Term("c", None)), Term("b", None))
        c4 = Clause(Term("a", Term("c", None)), Term("b", None))
        c5 = Clause(Term("a", None), Term("b", None))
        c6 = Clause(Term("c", None), Term("b", None))
        c7 = Clause(Term("a", Var("_")), Term("b", None))
        c8 = Clause(Term("a", Var("X1")), Term("b", Var("X1")))
        c9 = Clause(Term("a", Var("X2")), Term("b", Var("X2")))
        c10 = Clause(Term("a", -1), Term("b", -1))
        c11 = Clause(Term("a", -2), Term("b", -2))
        c12 = Clause(Term("a", -1), Term("b", -2))
        self.assertTrue(c1 == c2)
        self.assertTrue(c3 == c4)
        self.assertFalse(c5 == c6)
        self.assertFalse(c7 == c1)
        self.assertFalse(c8 == c9)
        self.assertFalse(c1 == c10)
        self.assertFalse(c8 == c10)
        self.assertFalse(c1 == c11)
        self.assertFalse(c1 == c12)
        self.assertFalse(c10 == c11)
        self.assertFalse(c10 == c12)

    def test_from_string(self):
        c1 = Term.from_string("a.")
        c2 = Term("a")
        c3 = Term.from_string("a:-b.")
        c4 = Clause(Term("a"), Term("b"))
        c5 = Term.from_string("a")
        self.assertTrue(c1 == c2)
        self.assertTrue(c3 == c4)
        self.assertTrue(c1 == c5)
        self.assertTrue(c2 == c5)

    def test_from_list(self):
        c1 = Term("a") | Term("b")
        c2 = Or.from_list([Term("a"), Term("b")])
        c3 = Term("a") | (Term("b") | Term("c"))
        c4 = Or.from_list([Term("a"), Term("b"), Term("c")])
        c5 = Or.from_list([])
        c6 = Term("fail")
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 == c4)
        self.assertTrue(c3 == c4)
        self.assertTrue(c5 == c6)

        c1 = Term("a") & Term("b")
        c2 = And.from_list([Term("a"), Term("b")])
        c3 = Term("a") & (Term("b") & Term("c"))
        c4 = And.from_list([Term("a"), Term("b"), Term("c")])
        c5 = And.from_list([])
        c6 = Term("true")
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 == c4)
        self.assertTrue(c3 == c4)
        self.assertTrue(c5 == c6)

    def test_to_list(self):
        term_list = [Term("a"), Term("b"), Term("c")]
        c1 = And.from_list(term_list)
        self.assertTrue(term_list == c1.to_list())
        self.assertTrue([Term("a"), Term("b"), Term("c")] == c1.to_list())

        c2 = Or.from_list(term_list)
        self.assertTrue(term_list == c2.to_list())
        self.assertTrue([Term("a"), Term("b"), Term("c")] == c2.to_list())

    def test_annotated_disjunction(self):
        c1 = AnnotatedDisjunction([Term("a")], None)
        c2 = AnnotatedDisjunction([Term("a")], None)
        c3 = AnnotatedDisjunction([Term("b")], None)
        c4 = AnnotatedDisjunction([Term("a"), Term("b")], Term("c"))
        c5 = AnnotatedDisjunction([Term("a"), Term("b")], Term("c"))
        c6 = AnnotatedDisjunction([Term("a"), Term("b")], Term("d"))
        c7 = AnnotatedDisjunction([Term("a"), Term("b")], None)
        self.assertTrue(c1 == c2)
        self.assertFalse(c1 == c3)
        self.assertTrue(c4 == c5)
        self.assertFalse(c4 == c6)
        self.assertFalse(c4 == c7)



#! /usr/bin/env python

import sys
import os


version_file = os.path.join(
    os.path.abspath(os.path.dirname(__file__)), "problog/version.py"
)

version = {}
with open(version_file) as fp:
    exec(fp.read(), version)
version = version["version"]

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

if __name__ == "__main__" and len(sys.argv) == 1:
    from problog import setup as problog_setup

    problog_setup.install()
elif __name__ == "__main__":

    from setuptools import setup, find_packages
    from setuptools.command.install import install

    class ProbLogInstall(install):
        def run(self):
            install.run(self)
            before_dir = os.getcwd()
            sys.path.insert(0, self.install_lib)
            from problog import setup as problog_setup

            try:
                problog_setup.install()
            except Exception as err:
                print("Optional ProbLog installation failed: %s" % err, file=sys.stderr)
            os.chdir(before_dir)

    package_data = {
        "problog": [
            "bin/darwin/cnf2dDNNF_wine",
            "bin/darwin/dsharp",
            "bin/darwin/maxsatz",
            "bin/linux/dsharp",
            "bin/linux/maxsatz",
            "bin/source/maxsatz/maxsatz2009.c",
            "bin/windows/dsharp.exe",
            "bin/windows/maxsatz.exe",
            "bin/windows/libgcc_s_dw2-1.dll",
            "bin/windows/libstdc++-6.dll",
            "web/*.py",
            "web/editor_local.html" "web/editor_adv.html",
            "web/js/problog_editor.js",
            "library/*.pl",
            "library/*.py",
            "library/nlp4plp.d/*",
        ]
    }

    setup(
        name="problog",
        version=version,
        description="ProbLog2: Probabilistic Logic Programming toolbox",
        long_description=long_description,
        long_description_content_type='text/markdown',
        url="https://dtai.cs.kuleuven.be/problog",
        author="ProbLog team",
        author_email="anton.dries@cs.kuleuven.be",
        license="Apache Software License",
        classifiers=[
            "Development Status :: 4 - Beta",
            "License :: OSI Approved :: Apache Software License",
            "Intended Audience :: Science/Research",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10",
            "Programming Language :: Python :: 3.11",
            "Programming Language :: Python :: 3.12",
            "Programming Language :: Prolog",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
        ],
        keywords="prolog probabilistic logic",
        packages=find_packages(),
        install_requires = ["setuptools"],
        extras_require={"sdd": ["pysdd>=0.2.6"]},
        entry_points={"console_scripts": ["problog=problog.tasks:main"]},
        package_data=package_data,
        cmdclass={"install": ProbLogInstall},
    )


def increment_release(v):
    v = v.split(".")
    if len(v) == 4:
        v = v[:3] + [str(int(v[3]) + 1)]
    else:
        v = v[:4]
    return ".".join(v)


def increment_dev(v):
    v = v.split(".")
    if len(v) == 4:
        v = v[:3] + [str(int(v[3]) + 1), "dev1"]
    else:
        v = v[:4] + ["dev" + str(int(v[4][3:]) + 1)]
    return ".".join(v)


def increment_version_dev():
    v = increment_dev(version)
    os.path.dirname(__file__)
    with open(version_file, "w") as f:
        f.write("version = '%s'\n" % v)


def increment_version_release():
    v = increment_release(version)
    with open(version_file, "w") as f:
        f.write("version = '%s'\n" % v)



from problog.evaluator import Semiring
from problog.extern import problog_export


@problog_export("+term", "+term", "+term", "+term", "+term", "+str", "+str")
def use_semiring(plus, times, zero, one, neg, is_dsp, is_nsp):
    class CustomSemiring(Semiring):
        def is_one(self, value):
            return value == self.one()

    CustomSemiring.zero = problog_export.database.create_function(zero, 1)
    CustomSemiring.one = problog_export.database.create_function(one, 1)
    CustomSemiring.plus = problog_export.database.create_function(plus, 3)
    CustomSemiring.times = problog_export.database.create_function(times, 3)
    CustomSemiring.pos_value = lambda s, x, k: x
    CustomSemiring.neg_value = problog_export.database.create_function(neg, 2)
    CustomSemiring.is_dsp = lambda s: is_dsp == "true"
    CustomSemiring.is_nsp = lambda s: is_nsp == "true"

    semiring = CustomSemiring()
    problog_export.database.set_data("semiring", semiring)

    return ()


@problog_export("+term", "+term", "+term", "+term", "+term", "+term", "+str", "+str")
def use_semiring(plus, times, zero, one, pos, neg, is_dsp, is_nsp):
    class CustomSemiring(Semiring):
        def is_one(self, value):
            return value == self.one()

    CustomSemiring.zero = problog_export.database.create_function(zero, 1)
    CustomSemiring.one = problog_export.database.create_function(one, 1)
    CustomSemiring.plus = problog_export.database.create_function(plus, 3)
    CustomSemiring.times = problog_export.database.create_function(times, 3)
    CustomSemiring.pos_value = problog_export.database.create_function(pos, 3)
    CustomSemiring.neg_value = problog_export.database.create_function(neg, 3)
    CustomSemiring.is_dsp = lambda s: is_dsp == "true"
    CustomSemiring.is_nsp = lambda s: is_nsp == "true"

    semiring = CustomSemiring()
    problog_export.database.set_data("semiring", semiring)

    return ()


@problog_export(
    "+term", "+term", "+term", "+term", "+term", "+term", "+term", "+str", "+str"
)
def use_semiring(plus, times, zero, one, pos, neg, negate, is_dsp, is_nsp):
    class CustomSemiring(Semiring):
        def is_one(self, value):
            return value == self.one()

    CustomSemiring.zero = problog_export.database.create_function(zero, 1)
    CustomSemiring.one = problog_export.database.create_function(one, 1)
    CustomSemiring.plus = problog_export.database.create_function(plus, 3)
    CustomSemiring.times = problog_export.database.create_function(times, 3)
    CustomSemiring.pos_value = problog_export.database.create_function(pos, 3)
    CustomSemiring.neg_value = problog_export.database.create_function(neg, 3)
    CustomSemiring.is_dsp = lambda s: is_dsp == "true"
    CustomSemiring.is_nsp = lambda s: is_nsp == "true"
    CustomSemiring.negate = problog_export.database.create_function(negate, 2)

    semiring = CustomSemiring()
    problog_export.database.set_data("semiring", semiring)

    return ()



"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import glob
import os
import sys
import unittest

from problog import get_evaluatable, register_semiring
from problog import root_path
from problog.ddnnf_formula import DDNNF
from problog.evaluator import SemiringProbability, SemiringLogProbability, Semiring
from problog.formula import LogicFormula
from problog.forward import _ForwardSDD
from problog.program import PrologFile, DefaultPrologParser, ExtendedPrologFactory

if __name__ == "__main__":
    sys.path.insert(
        0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../../"))
    )

# noinspection PyBroadException
try:
    from pysdd import sdd

    has_sdd = True
except Exception as err:
    print("SDD library not available due to error: ", err, file=sys.stderr)
    has_sdd = False


class TestSystemSpecific(unittest.TestCase):
    def setUp(self):
        try:
            self.assertSequenceEqual = self.assertItemsEqual
        except AttributeError:
            self.assertSequenceEqual = self.assertCountEqual

    def test_keep_order(self):
        """

        :return:
        """
        filename = root_path("test/specific/", "bug_keep_order.pl")
        correct = read_result(filename)
        pl = PrologFile(filename)
        for avoid_name_clash in [True, False]:
            for keep_order in [True, False]:
                for label_all in [True, False]:
                    with self.subTest(
                        avoid_name_clash=avoid_name_clash,
                        keep_order=keep_order,
                        label_all=label_all,
                    ):
                        try:
                            lf = LogicFormula.create_from(
                                pl,
                                avoid_name_clash=avoid_name_clash,
                                keep_order=keep_order,
                                label_all=label_all,
                            )
                            ddnnf = DDNNF.create_from(lf)

                            computed = ddnnf.evaluate()
                            computed = {str(k): v for k, v in computed.items()}
                        except Exception as err:
                            print(err)
                            e = err
                            computed = None

                        if computed is None:
                            self.assertEqual(correct, type(e).__name__)
                        else:
                            self.assertIsInstance(correct, dict)
                            self.assertSequenceEqual(correct, computed)

                            for query in correct:
                                self.assertAlmostEqual(
                                    correct[query], computed[query], msg=query
                                )


class TestDummy(unittest.TestCase):
    def test_dummy(self):
        pass


class TestSystemGeneric(unittest.TestCase):
    def setUp(self):
        try:
            self.assertSequenceEqual = self.assertItemsEqual
        except AttributeError:
            self.assertSequenceEqual = self.assertCountEqual


def read_result(filename):
    results = {}
    with open(filename) as f:
        reading = False
        for l in f:
            l = l.strip()
            if l.startswith("%Expected outcome:"):
                reading = True
            elif reading:
                if l.lower().startswith("% error"):
                    return l[len("% error") :].strip()
                elif l.startswith("% "):
                    query, prob = l[2:].rsplit(None, 1)
                    results[query.strip()] = float(prob.strip())
                else:
                    reading = False
            if l.startswith("query(") and l.find("% outcome:") >= 0:
                pos = l.find("% outcome:")
                query = l[6:pos].strip().rstrip(".").rstrip()[:-1]
                prob = l[pos + 10 :]
                results[query.strip()] = float(prob.strip())
    return results


def createSystemTestGeneric(filename, logspace=False):
    correct = read_result(filename)

    def test(self):
        semirings = {
            "Default": None,
            "Custom": SemiringProbabilityCopy(),
            "CustomNSP": SemiringProbabilityNSPCopy(),
        }
        for eval_name in evaluatables:
            for semiring in semirings:
                with self.subTest(evaluatable_name=eval_name, semiring=semiring):
                    evaluate(
                        self,
                        evaluatable_name=eval_name,
                        custom_semiring=semirings[semiring],
                    )

        # explicit encoding from ForwardSDD
        if has_sdd:
            for semiring in semirings:
                with self.subTest(semiring=semiring):
                    evaluate_explicit_from_fsdd(
                        self, custom_semiring=semirings[semiring]
                    )

    def evaluate(self, evaluatable_name=None, custom_semiring=None):
        try:
            parser = DefaultPrologParser(ExtendedPrologFactory())
            kc = get_evaluatable(name=evaluatable_name).create_from(
                PrologFile(filename, parser=parser)
            )

            if custom_semiring is not None:
                semiring = custom_semiring  # forces the custom semiring code.
            elif logspace:
                semiring = SemiringLogProbability()
            else:
                semiring = SemiringProbability()

            computed = kc.evaluate(semiring=semiring)
            computed = {str(k): v for k, v in computed.items()}
        except Exception as err:
            # print("exception %s" % err)
            e = err
            computed = None

        if computed is None:
            self.assertEqual(correct, type(e).__name__)
        else:
            self.assertIsInstance(correct, dict)
            self.assertSequenceEqual(correct, computed)

            for query in correct:
                self.assertAlmostEqual(correct[query], computed[query], msg=query)

    def evaluate_explicit_from_fsdd(self, custom_semiring=None):
        try:
            parser = DefaultPrologParser(ExtendedPrologFactory())
            lf = PrologFile(filename, parser=parser)
            kc = _ForwardSDD.create_from(lf)  # type: _ForwardSDD
            kc = kc.to_explicit_encoding()

            if custom_semiring is not None:
                semiring = custom_semiring  # forces the custom semiring code.
            elif logspace:
                semiring = SemiringLogProbability()
            else:
                semiring = SemiringProbability()

            computed = kc.evaluate(semiring=semiring)
            computed = {str(k): v for k, v in computed.items()}
        except Exception as err:
            # print("exception %s" % err)
            e = err
            computed = None

        if computed is None:
            self.assertEqual(correct, type(e).__name__)
        else:
            self.assertIsInstance(correct, dict)
            self.assertSequenceEqual(correct, computed)

            for query in correct:
                self.assertAlmostEqual(correct[query], computed[query], msg=query)

    return test


class SemiringProbabilityCopy(Semiring):
    """Mocking SemiringProbability to force the 'custom semiring' code -> Must not extend SemiringProbability."""

    def __init__(self):
        self._semiring = SemiringProbability()

    def one(self):
        return self._semiring.one()

    def zero(self):
        return self._semiring.zero()

    def is_one(self, value):
        return self._semiring.is_one(value)

    def is_zero(self, value):
        return self._semiring.is_zero(value)

    def plus(self, a, b):
        return self._semiring.plus(a, b)

    def times(self, a, b):
        return self._semiring.times(a, b)

    def negate(self, a):
        return self._semiring.negate(a)

    def normalize(self, a, z):
        return self._semiring.normalize(a, z)

    def value(self, a):
        return self._semiring.value(a)

    def is_dsp(self):
        return self._semiring.is_dsp()

    def in_domain(self, a):
        return self._semiring.in_domain(a)

    @classmethod
    def create(cls, *, engine, database, **kwargs):
        return cls()


class SemiringProbabilityNSPCopy(SemiringProbabilityCopy):
    def is_nsp(self):
        return True

    def pos_value(self, a, key=None):
        if isinstance(a, tuple):
            return float(a[0])
        else:
            return float(a)

    def neg_value(self, a, key=None):
        if isinstance(a, tuple):
            return float(a[1])
        else:
            return 1 - float(a)


register_semiring("prob_copy", SemiringProbabilityCopy)
register_semiring("prob_nsp_copy", SemiringProbabilityNSPCopy)


if __name__ == "__main__":
    filenames = sys.argv[1:]
else:
    filenames = glob.glob(root_path("test", "*.pl"))

evaluatables = ["ddnnf"]

if has_sdd:
    evaluatables.append("sdd")
    evaluatables.append("sddx")
    evaluatables.append("fsdd")
else:
    print("No SDD support - The system tests are not performed with SDDs.")

for testfile in filenames:
    testname = "test_system_" + os.path.splitext(os.path.basename(testfile))[0]
    setattr(TestSystemGeneric, testname, createSystemTestGeneric(testfile, True))

if __name__ == "__main__":
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSystemGeneric)
    unittest.TextTestRunner(verbosity=2).run(suite)
    suite = unittest.TestLoader().loadTestsFromTestCase(TestSystemSpecific)
    unittest.TextTestRunner(verbosity=2).run(suite)



"""
Maximum A-Posteriori inference for ProbLog (MAP)
"""

import sys
import traceback

from .dtproblog import (
    search_exhaustive,
    search_local,
    argparser,
    print_result,
    print_result_json,
)
from .. import get_evaluatable
from ..constraint import TrueConstraint
from ..formula import LogicFormula
from ..program import PrologFile


def main(argv, result_handler=None):
    args = argparser().parse_args(argv)

    if result_handler is None:
        if args.web:
            result_handler = print_result_json
        else:
            result_handler = print_result

    if args.output is not None:
        outf = open(args.output, "w")
    else:
        outf = sys.stdout

    search = args.search

    final_result = None
    try:

        # certain queries are given, they refer to facts
        #  we want to find assignments for these
        #  such that the overall probability is maximized
        # query facts => decisions with utility log p

        # Load the model
        model = PrologFile(args.inputfile)

        # Ground the model
        ground_program = LogicFormula.create_from(model)

        # Get the conditional probabilities of the query facts
        query_probs = get_evaluatable().create_from(ground_program).evaluate()

        # Extract queries and replace them with decisions
        decisions = []
        utilities = []
        decision_nodes = set()
        for qn, qi, ql in ground_program.labeled():
            node = ground_program.get_node(qi)
            prob = query_probs[qn]
            if type(node).__name__ != "atom":
                raise Exception("Queries should be facts: '%s'" % qn)
            decisions.append((qi, qn))
            decision_nodes.add(qi)
            probability = prob
            utilities.append((qn, float(probability)))
            utilities.append((-qn, 1 - float(probability)))
            # utilities.append((qn, math.log(float(probability)) - math.log(1 - float(probability))))
        utilities = dict(utilities)

        # Add constraints that enforce the evidence
        for qn, qi in ground_program.evidence():
            ground_program.add_constraint(TrueConstraint(qi))
        ground_program.clear_evidence()

        # Process constraints (only on decisions, the others are encoded in the model)
        constraints = []
        for c in ground_program.constraints():
            if set(c.get_nodes()) & decision_nodes:
                constraints.append(c)

        # Compile the program (again)
        knowledge = get_evaluatable().create_from(ground_program)

        # Use dt-problog search to find the solution
        if search == "local":
            result = search_local(knowledge, decisions, utilities, constraints)
        else:
            result = search_exhaustive(knowledge, decisions, utilities, constraints)

        # Print the result
        decisions, score, stats = result

        # score = math.exp(-score)
        final_result = (True, (decisions, score, stats))
    except Exception as err:
        err.trace = traceback.format_exc()
        final_result = (False, err)

    result_handler(final_result, outf)

    if args.output is not None:
        outf.close()

    return final_result



from collections import OrderedDict

from problog.extern import problog_export, problog_export_nondet
from problog.logic import Constant

recdb_key = "_rec_db"


class LL(object):
    def __init__(self):
        self.first = None
        self.last = None

    def push_left(self, item):
        entry = LLEntry(self, item, None, self.first)
        if self.first is None:  # list is currently empty
            self.first = entry
            self.last = entry
        else:
            self.first.prv = entry
            self.first = entry
        return entry

    def push_right(self, item):
        entry = LLEntry(self, item, self.last, None)
        if self.last is None:  # list if currently empty
            self.first = entry
            self.last = entry
        else:
            self.last.nxt = entry
            self.last = entry
        return entry

    def erase(self, item):
        if item.prv is None and item.nxt is None:  # only element in list
            self.first = None
            self.last = None
        elif item.prv is None:  # first element
            item.nxt.prv = None
            self.first = item.nxt
        elif item.nxt is None:  # last element
            item.prv.nxt = None
            self.last = item.prv
        else:
            item.prv.nxt = item.nxt
            item.nxt.prv = item.prv

    def is_empty(self):
        return self.first is None

    def __iter__(self):
        it = self.first
        while it is not None:
            yield it
            it = it.nxt


class LLEntry(object):
    __slots__ = ("lst", "val", "prv", "nxt")

    def __init__(self, lst, val, prv, nxt):
        self.lst = lst
        self.prv = prv
        self.nxt = nxt
        self.val = val

    def erase(self):
        self.lst.erase(self)


def record_base(key, term, append):
    db = problog_export.database.get_data(recdb_key)
    if db is None:
        db = OrderedDict()
        problog_export.database.set_data(recdb_key, db)
    lst = db.get(key)
    if lst is None:
        lst = LL()
        db[key] = lst
    if append:
        return lst.push_right(term)
    else:
        return lst.push_left(term)


@problog_export_nondet("-term")
def current_key():
    db = problog_export.database.get_data(recdb_key, {})
    return list(db.keys())


@problog_export("+term", "+term", "-term")
def recorda(key, term):
    return Constant(record_base(key, term, False))


@problog_export("+term", "+term")
def recorda(key, term):
    record_base(key, term, False)
    return ()


@problog_export("+term", "+term", "-int")
def recordz(key, term):
    return Constant(record_base(key, term, True))


@problog_export("+term", "+term")
def recordz(key, term):
    record_base(key, term, True)
    return ()


@problog_export("+term")
def erase(ref):
    ref.functor.erase()
    return ()


@problog_export_nondet("+term", "-term")
def recorded(key):
    db = problog_export.database.get_data(recdb_key, {})
    lst = db.get(key)
    if lst is None:
        return ()
    else:
        return [x.val for x in lst]


@problog_export_nondet("+term", "-term", "-term")
def recorded(key):
    db = problog_export.database.get_data(recdb_key, {})
    lst = db.get(key)
    if lst is None:
        return ()
    else:
        return [(x.val, Constant(x)) for x in lst]


@problog_export("+term", "-term")
def instance(ref):
    return ref.functor.val


# @problog_export_nondet('-term', '-term', '-term')
# def recorded():
#     db = problog_export.database.get_data(recdb_key)
#
#     for key, lst in db.items():
#         if lst is None:
#             return ()
#         else:
#             return [(key, x.val, Constant(x)) for x in lst]



"""
problog.engine - Grounding engine
---------------------------------

Grounding engine to transform a ProbLog program into a propositional formula.

..
    Part of the ProbLog distribution.

    Copyright 2015 KU Leuven, DTAI Research Group

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
import logging
from collections import defaultdict
from subprocess import CalledProcessError

from .core import transform
from .engine_unify import *
from .errors import NonGroundQuery
from .formula import LogicFormula
from .logic import *
from .program import LogicProgram
from .util import Timer


@transform(LogicProgram, LogicFormula)
def ground(model, target=None, grounder=None, **kwdargs):
    """Ground a given model.

    :param model: logic program to ground
    :type model: LogicProgram
    :return: the ground program
    :rtype: LogicFormula
    """
    if grounder in ("yap", "yap_debug"):
        from .ground_yap import ground_yap

        return ground_yap(model, target, **kwdargs)
    else:
        return ground_default(model, target, **kwdargs)


@transform(LogicProgram, LogicFormula)
def ground_default(
    model,
    target=None,
    queries=None,
    evidence=None,
    propagate_evidence=False,
    labels=None,
    engine=None,
    **kwdargs
):
    """Ground a given model.

    :param model: logic program to ground
    :type model: LogicProgram
    :param target: formula in which to store ground program
    :type target: LogicFormula
    :param queries: list of queries to override the default
    :param evidence: list of evidence atoms to override the default
    :return: the ground program
    :rtype: LogicFormula
    """
    if engine is None:
        engine = DefaultEngine(**kwdargs)
    return engine.ground_all(
        model,
        target,
        queries=queries,
        evidence=evidence,
        propagate_evidence=propagate_evidence,
        labels=labels,
    )


class GenericEngine(object):  # pragma: no cover
    """Generic interface to a grounding engine."""

    def prepare(self, db: LogicProgram):
        """Prepare the given database for querying.
        Calling this method is optional.

        :param db: logic program
        :returns: logic program in optimized format where builtins are initialized and directives \
        have been evaluated
        """
        raise NotImplementedError("GenericEngine.prepare is an abstract method.")

    def query(self, db: LogicProgram, term):
        """Evaluate a query without generating a ground program.

        :param db: logic program
        :param term: term to query; variables should be represented as None
        :returns: list of tuples of argument for which the query succeeds.
        """
        raise NotImplementedError("GenericEngine.query is an abstract method.")

    def ground(self, db: LogicProgram, term, target=None, label=None):
        """Ground a given query term and store the result in the given ground program.

        :param db: logic program
        :param term: term to ground; variables should be represented as None
        :param target: target logic formula to store grounding in (a new one is created if none is \
        given)
        :param label: optional label (query, evidence, ...)
        :returns: logic formula (target if given)
        """
        raise NotImplementedError("GenericEngine.ground is an abstract method.")

    def ground_all(self, db: LogicProgram, target=None, queries=None, evidence=None):
        """Ground all queries and evidence found in the the given database.

        :param db: logic program
        :param target: logic formula to ground into
        :param queries: list of queries to evaluate instead of the ones in the logic program
        :param evidence: list of evidence to evaluate instead of the ones in the logic program
        :returns: ground program
        """
        raise NotImplementedError("GenericEngine.ground_all is an abstract method.")


class ClauseDBEngine(GenericEngine):
    """Parent class for all Python ClauseDB-based engines."""

    UNKNOWN_ERROR = 0
    UNKNOWN_FAIL = 1

    def __init__(self, builtins=True, **kwdargs):
        self.__builtin_index = {}
        self.__builtins = []
        self.__externals = {}

        self._unique_number = 0
        self.unknown = kwdargs.get("unknown", self.UNKNOWN_ERROR)

        if builtins:
            self.load_builtins()

        self.functions = {}
        self.args = kwdargs.get("args")

    def load_builtins(self):
        """Load default builtins."""
        raise NotImplementedError(
            "ClauseDBEngine.loadBuiltIns is an abstract function."
        )

    def get_builtin(self, index):
        """Get builtin's evaluation function based on its identifier.
       :param index: index of the builtin
       :return: function that evaluates the builtin
        """
        real_index = -(index + 1)
        return self.__builtins[real_index]

    def add_builtin(self, predicate, arity, function):
        """Add a builtin.

        :param predicate: name of builtin predicate
        :param arity: arity of builtin predicate
        :param function: function to execute builtin
        """
        sig = "%s/%s" % (predicate, arity)
        self.__builtin_index[sig] = -(len(self.__builtins) + 1)
        self.__builtins.append(function)

    def get_builtins(self):
        """Get the list of builtins."""
        return self.__builtin_index

    def prepare(self, db):
        """Convert given logic program to suitable format for this engine.
        Calling this method is optional, but it allows to perform multiple operations on the same \
        database.
        This also executes any directives in the input model.

        :param db: logic program to prepare for evaluation
        :return: logic program in a suitable format for this engine
        :rtype: ClauseDB
        """
        result = ClauseDB.createFrom(db, builtins=self.get_builtins())
        result.engine = self
        self._process_directives(result)
        return result

    def call(self, query, database, target, transform=None, **kwdargs):
        raise NotImplementedError("ClauseDBEngine.call is an abstract function.")

    def execute(self, node_id, database=None, context=None, target=None, **kwdargs):
        raise NotImplementedError("ClauseDBEngine.execute is an abstract function.")

    def get_non_cache_functor(self):
        """Get a unique functor that is excluded from caching.

        :return: unique functor that is excluded from caching
        :rtype: basestring
        """
        self._unique_number += 1
        return "_nocache_%s" % self._unique_number

    def _process_directives(self, db):
        """Process directives present in the database."""
        term = Term("_directive")
        directive_node = db.find(term)
        if directive_node is not None:
            directives_list = db.get_node(directive_node).children
            # Emulate the order in which directives are encountered using a stack.
            pending_directives = []
            while directives_list:
                pending_directives.append(directives_list.pop())

            gp = LogicFormula()
            while pending_directives:
                current = pending_directives.pop()
                self.execute(
                    current,
                    database=db,
                    context=self.create_context((), define=None),
                    target=gp,
                )
                while directives_list:
                    pending_directives.append(directives_list.pop())

        return True

    # noinspection PyUnusedLocal
    def create_context(self, content, define=None, parent=None):
        """Create a variable context."""
        return content

    def _fix_context(self, context):
        return tuple(context)

    def _clone_context(self, context):
        return list(context)

    def query(self, db, term, backend=None, **kwdargs):
        """

        :param db:
        :param term:
        :param kwdargs:
        :return:
        """

        if backend in ("swipl", "yap"):
            from .util import mktempfile, subprocess_check_output

            tmpfn = mktempfile(".pl")
            with open(tmpfn, "w") as tmpf:
                print(db.to_prolog(), file=tmpf)

            from problog.logic import term2str

            termstr = term2str(term)
            cmd = [
                "swipl",
                "-l",
                tmpfn,
                "-g",
                "%s, writeln(%s), fail; halt" % (termstr, termstr),
            ]

            try:
                output = subprocess_check_output(cmd)
            except CalledProcessError as err:
                in_error = True
                error_message = []
                for line in err.output.split("\n"):
                    if line.startswith("Warning:"):
                        in_error = False
                    elif line.startswith("ERROR:"):
                        in_error = True
                    if in_error:
                        error_message.append(line)
                error_message = "SWI-Prolog returned some errors:\n" + "\n".join(
                    error_message
                )
                raise GroundingError(error_message)

            return [
                Term.from_string(line).args
                for line in output.split("\n")
                if line.strip()
            ]
        else:
            gp = LogicFormula()
            if term.is_negated():
                term = -term
                negative = True
            else:
                negative = False
            gp, result = self._ground(db, term, gp, **kwdargs)
            if negative:
                if not result:
                    return [term]
                else:
                    return []
            else:
                return [x for x, y in result]

    def ground(self, db, term, target=None, label=None, **kwdargs):
        """Ground a query on the given database.

        :param db: logic program
        :type db: LogicProgram
        :param term: query term
        :type term: Term
        :param label: type of query (e.g. ``query``, ``evidence`` or ``-evidence``)
        :type label: str
        :param kwdargs: additional arguments
        :return: ground program containing the query
        :rtype: LogicFormula
        """
        if term.is_negated():
            negated = True
            term = -term
        elif term.functor in ("not", "\\+") and term.arity == 1:
            negated = True
            term = term.args[0]
        else:
            negated = False

        target, results = self._ground(db, term, target, silent_fail=False, **kwdargs)

        args_node = defaultdict(list)
        for args, node_id in results:
            if not is_ground(*args) and target.is_probabilistic(node_id):
                raise NonGroundQuery(term, db.lineno(term.location))
            args = tuple(args)
            args_node[args].append(node_id)
        for args, node_ids in args_node.items():
            if len(node_ids) > 1:
                node_id = target.add_or(node_ids)
            else:
                node_id = node_ids[0]
            term_store = term.with_args(*args)
            if negated:
                target.add_name(-term_store, target.negate(node_id), label)
            else:
                target.add_name(term_store, node_id, label)
        if not results:
            if negated:
                target.add_name(-term, target.TRUE, label)
            else:
                target.add_name(term, target.FALSE, label)

        return target

    def ground_step(
        self, db, term, gp=None, silent_fail=True, assume_prepared=False, **kwdargs
    ):
        """

        :param db:
        :type db: LogicProgram
        :param term:
        :param gp:
        :param silent_fail:
        :param assume_prepared:
        :param kwdargs:
        :return:
        """
        # Convert logic program if needed.
        if not assume_prepared:
            db = self.prepare(db)
        # Create a new target datastructure if none was given.
        if gp is None:
            gp = LogicFormula()
        # Find the define node for the given query term.
        clause_node = db.find(term)
        # If term not defined: fail query (no error)    # TODO add error to make it consistent?
        if clause_node is None:
            # Could be builtin?
            clause_node = db.get_builtin(term.signature)
        if clause_node is None:
            if silent_fail or self.unknown == self.UNKNOWN_FAIL:
                return []
            else:
                raise UnknownClause(term.signature, location=db.lineno(term.location))

        try:
            term = term.apply(_ReplaceVar())  # replace Var(_) by integers

            context = self.create_context(term.args)
            context, xxx = substitute_call_args(context, context, 0)
            actions = self.execute_init(
                clause_node, database=db, target=gp, context=context, **kwdargs
            )
        except UnknownClauseInternal:
            if silent_fail or self.unknown == self.UNKNOWN_FAIL:
                return []
            else:
                raise UnknownClause(term.signature, location=db.lineno(term.location))
        return actions

    def _ground(
        self, db, term, gp=None, silent_fail=True, assume_prepared=False, **kwdargs
    ):
        """

        :param db:
        :type db: LogicProgram
        :param term:
        :param gp:
        :param silent_fail:
        :param assume_prepared:
        :param kwdargs:
        :return:
        """
        # Convert logic program if needed.
        if not assume_prepared:
            db = self.prepare(db)
        # Create a new target datastructure if none was given.
        if gp is None:
            gp = LogicFormula()
        # Find the define node for the given query term.
        clause_node = db.find(term)
        # If term not defined: fail query (no error)    # TODO add error to make it consistent?
        if clause_node is None:
            # Could be builtin?
            clause_node = db.get_builtin(term.signature)
        if clause_node is None:
            if silent_fail or self.unknown == self.UNKNOWN_FAIL:
                return gp, []
            else:
                raise UnknownClause(term.signature, location=db.lineno(term.location))

        try:
            term = term.apply(_ReplaceVar())  # replace Var(_) by integers

            context = self.create_context(term.args)
            context, xxx = substitute_call_args(context, context, 0)
            if self.debugger:
                location = db.lineno(term.location)
                self.debugger.call_create(
                    clause_node, term.functor, context, None, location
                )
            results = self.execute(
                clause_node, database=db, target=gp, context=context, **kwdargs
            )
        except UnknownClauseInternal:
            if silent_fail or self.unknown == self.UNKNOWN_FAIL:
                return gp, []
            else:
                raise UnknownClause(term.signature, location=db.lineno(term.location))
        return gp, results

    def ground_evidence(self, db, target, evidence, propagate_evidence=False):
        logger = logging.getLogger("problog")
        # Ground evidence
        for query in evidence:
            if len(query) == 1:  # evidence/1
                if query[0].is_negated():
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(
                        db,
                        -query[0],
                        target,
                        label=target.LABEL_EVIDENCE_NEG,
                        is_root=True,
                    )
                    logger.debug("Ground program size: %s", len(target))
                else:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(
                        db,
                        query[0],
                        target,
                        label=target.LABEL_EVIDENCE_POS,
                        is_root=True,
                    )
                    logger.debug("Ground program size: %s", len(target))
            else:  # evidence/2
                if str(query[1]) == "true" or query[1] == True:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(
                        db,
                        query[0],
                        target,
                        label=target.LABEL_EVIDENCE_POS,
                        is_root=True,
                    )
                    logger.debug("Ground program size: %s", len(target))
                elif str(query[1]) == "false" or query[1] == False:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(
                        db,
                        query[0],
                        target,
                        label=target.LABEL_EVIDENCE_NEG,
                        is_root=True,
                    )
                    logger.debug("Ground program size: %s", len(target))
                else:
                    logger.debug("Grounding evidence '%s'", query[0])
                    target = self.ground(
                        db,
                        query[0],
                        target,
                        label=target.LABEL_EVIDENCE_MAYBE,
                        is_root=True,
                    )
                    logger.debug("Ground program size: %s", len(target))
        if propagate_evidence:
            with Timer("Propagating evidence"):
                target.lookup_evidence = {}
                ev_nodes = [
                    node
                    for name, node in target.evidence()
                    if node != 0 and node is not None
                ]
                target.propagate(ev_nodes, target.lookup_evidence)

    def ground_queries(self, db, target, queries):
        logger = logging.getLogger("problog")
        for label, query in queries:
            logger.debug("Grounding query '%s'", query)
            target = self.ground(db, query, target, label=label)
            logger.debug("Ground program size: %s", len(target))

    def ground_all(
        self,
        db,
        target=None,
        queries=None,
        evidence=None,
        propagate_evidence=False,
        labels=None,
    ):
        if labels is None:
            labels = []
        # Initialize target if not given.
        if target is None:
            target = LogicFormula()

        db = self.prepare(db)

        logger = logging.getLogger("problog")
        with Timer("Grounding"):
            # Load queries: use argument if available, otherwise load from database.
            if queries is None:
                queries = [q[0] for q in self.query(db, Term("query", None))]
            for query in queries:
                if not isinstance(query, Term):
                    raise GroundingError("Invalid query")  # TODO can we add a location?
            # Load evidence: use argument if available, otherwise load from database.
            if evidence is None:
                evidence = self.query(db, Term("evidence", None, None))
                evidence += self.query(db, Term("evidence", None))

            queries = [(target.LABEL_QUERY, q) for q in queries]
            for label, arity in labels:
                queries += [
                    (label, q[0])
                    for q in self.query(db, Term(label, *([None] * arity)))
                ]

            for ev in evidence:
                if not isinstance(ev[0], Term):
                    raise GroundingError(
                        "Invalid evidence"
                    )  # TODO can we add a location?
            # Ground queries
            if propagate_evidence:
                self.ground_evidence(
                    db, target, evidence, propagate_evidence=propagate_evidence
                )
                self.ground_queries(db, target, queries)
                if hasattr(target, "lookup_evidence"):
                    logger.debug(
                        "Propagated evidence: %s" % list(target.lookup_evidence)
                    )
            else:
                self.ground_queries(db, target, queries)
                self.ground_evidence(db, target, evidence)
        return target

    def add_external_calls(self, externals):
        self.__externals.update(externals)

    def get_external_call(self, func_name):
        if self.__externals is None or func_name not in self.__externals:
            return None
        return self.__externals[func_name]


class _ReplaceVar(object):
    def __init__(self):
        self.translate = {}

    def __getitem__(self, name):
        if type(name) == str:
            if name in self.translate:
                return self.translate[name]
            else:
                v = -len(self.translate) - 1
                self.translate[name] = v
                return v
        else:
            return name


class UnknownClauseInternal(Exception):
    """Undefined clause in call used internally."""

    pass


class NonGroundProbabilisticClause(GroundingError):
    """Encountered a non-ground probabilistic clause."""

    def __init__(self, location):
        GroundingError.__init__(
            self, "Encountered a non-ground probabilistic clause", location
        )


class UnknownClause(GroundingError):
    """Undefined clause in call."""

    def __init__(self, signature, location):
        GroundingError.__init__(self, "No clauses found for '%s'" % signature, location)
        self.signature = signature


from .engine_stack import StackBasedEngine as DefaultEngine

from .clausedb import ClauseDB



from problog.extern import problog_export, problog_export_raw


@problog_export_raw("+term")
def assertz(term, target=None, **kwargs):
    problog_export.database += term
    target._cache.reset()  # reset tabling cache
    return [(term,)]


@problog_export_raw("+term")
def retract(term, target=None, **kwargs):
    db = problog_export.database
    nodekey = db.find(term)
    node = db.get_node(nodekey)
    to_erase = node.children.find(term.args)
    if to_erase:
        item = next(to_erase.__iter__())
        node.children.erase((item,))
        target._cache.reset()  # reset tabling cache
        return [(term,)]
    else:
        return []


@problog_export_raw("+term")
def retractall(term, target=None, **kwargs):
    db = problog_export.database
    nodekey = db.find(term)
    node = db.get_node(nodekey)
    to_erase = node.children.find(term.args)
    node.children.erase(to_erase)
    target._cache.reset()  # reset tabling cache
    return [(term,)]
